"""
Defines a tensorial field of rank 2 over a grid

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from typing import TYPE_CHECKING, Callable, Optional, Tuple, Union

import numba as nb
import numpy as np

from ..tools.docstrings import fill_in_docstring
from ..tools.misc import get_common_dtype
from ..tools.numba import get_common_numba_dtype, jit
from .base import DataFieldBase
from .scalar import ScalarField
from .vectorial import VectorField

if TYPE_CHECKING:
    from ..grids.boundaries.axes import BoundariesData  # @UnusedImport


class Tensor2Field(DataFieldBase):
    """Single tensor field of rank 2 on a grid

    Attributes:
        grid (:class:`~pde.grids.GridBase`):
            The underlying grid defining the discretization
        data (:class:`~numpy.ndarray`):
            Tensor components at the support points of the grid
        label (str):
            Name of the field
    """

    rank = 2

    def __getitem__(self, key: Tuple[int, int]) -> ScalarField:
        """ extract a component of the VectorField """
        try:
            if len(key) != 2:
                raise IndexError("Index must be given as two integers")
        except TypeError:
            raise IndexError("Index must be given as two integers")
        return ScalarField(self.grid, self.data[key])

    @DataFieldBase._data_flat.setter  # type: ignore
    def _data_flat(self, value):
        """ set the data from a value from a collection """
        dim = self.grid.dim
        self._data = value.reshape(dim, dim, *self.grid.shape)
        # check whether both point to the same memory location
        addr_value = value.__array_interface__["data"][0]
        addr_self_data = self._data.__array_interface__["data"][0]
        assert addr_value == addr_self_data

    def dot(
        self,
        other: Union[VectorField, "Tensor2Field"],
        out: Optional[Union[VectorField, "Tensor2Field"]] = None,
        *,
        conjugate: bool = True,
        label: str = "dot product",
    ) -> Union[VectorField, "Tensor2Field"]:
        """calculate the dot product involving a tensor field

        This supports the dot product between two tensor fields as well as the
        product between a tensor and a vector. The resulting fields will be a
        tensor or vector, respectively.

        Args:
            other (VectorField or Tensor2Field):
                the second field
            out (VectorField or Tensor2Field, optional):
                Optional field to which the  result is written.
            conjugate (bool):
                Whether to use the complex conjugate for the second operand
            label (str, optional):
                Name of the returned field

        Returns:
            VectorField or Tensor2Field: the result of applying the dot operator
        """
        # check input
        self.grid.assert_grid_compatible(other.grid)
        if not isinstance(other, (VectorField, Tensor2Field)):
            raise TypeError("Second term must be a vector or tensor field")

        # create and check the output instance
        if out is None:
            out = other.__class__(self.grid, dtype=get_common_dtype(self, other))
        else:
            assert isinstance(out, other.__class__), f"`out` must be {other.__class__}"
            self.grid.assert_grid_compatible(out.grid)

        # calculate the result
        other_data = other.data.conjugate() if conjugate else other.data
        np.einsum("ij...,j...->i...", self.data, other_data, out=out.data)
        if label is not None:
            out.label = label

        return out

    __matmul__ = dot  # support python @-syntax for matrix multiplication

    def make_dot_operator(
        self, backend: str = "numba", *, conjugate: bool = True
    ) -> Callable[[np.ndarray, np.ndarray, Optional[np.ndarray]], np.ndarray]:
        """return operator calculating the dot product involving vector fields

        This supports both products between two vectors as well as products
        between a vector and a tensor.

        Warning:
            This function does not check types or dimensions.

        Args:
            conjugate (bool):
                Whether to use the complex conjugate for the second operand

        Returns:
            function that takes two instance of :class:`~numpy.ndarray`, which
            contain the discretized data of the two operands. An optional third
            argument can specify the output array to which the result is
            written. Note that the returned function is jitted with numba for
            speed.
        """
        dim = self.grid.dim

        if backend == "numba":
            # create the dot product using a numba compiled function

            if conjugate:
                # create inner function calculating the dot product using conjugate

                @jit
                def calc(a: np.ndarray, b: np.ndarray, out: np.ndarray) -> np.ndarray:
                    """ calculate dot product between fields `a` and `b` """
                    for i in range(dim):
                        out[i] = a[i, 0] * b[0].conjugate()  # overwrite data in out
                        for j in range(1, dim):
                            out[i] += a[i, j] * b[j].conjugate()
                    return out

            else:
                # create the inner function calculating the dot product

                @jit
                def calc(a: np.ndarray, b: np.ndarray, out: np.ndarray) -> np.ndarray:
                    """ calculate dot product between fields `a` and `b` """
                    for i in range(dim):
                        out[i] = a[i, 0] * b[0]  # overwrite potential data in out
                        for j in range(1, dim):
                            out[i] += a[i, j] * b[j]
                    return out

            # build the outer function with the correct signature
            if nb.config.DISABLE_JIT:

                def dot(
                    a: np.ndarray, b: np.ndarray, out: np.ndarray = None
                ) -> np.ndarray:
                    """wrapper deciding whether the underlying function is called
                    with or without `out`."""
                    if out is None:
                        out = np.empty(b.shape, dtype=get_common_dtype(a, b))
                    return calc(a, b, out)  # type: ignore

            else:

                @nb.generated_jit
                def dot(
                    a: np.ndarray, b: np.ndarray, out: np.ndarray = None
                ) -> np.ndarray:
                    """wrapper deciding whether the underlying function is called
                    with or without `out`."""
                    if isinstance(a, nb.types.Number):
                        # simple scalar call -> do not need to allocate anything
                        raise RuntimeError("Dot needs to be called with fields")

                    elif isinstance(out, (nb.types.NoneType, nb.types.Omitted)):
                        # function is called without `out`
                        dtype = get_common_numba_dtype(a, b)

                        def f_with_allocated_out(
                            a: np.ndarray, b: np.ndarray, out: np.ndarray
                        ) -> np.ndarray:
                            """ helper function allocating output array """
                            out = np.empty(b.shape, dtype=dtype)
                            return calc(a, b, out=out)  # type: ignore

                        return f_with_allocated_out  # type: ignore

                    else:
                        # function is called with `out` argument
                        return calc  # type: ignore

        elif backend == "numpy":
            # create the dot product using basic numpy functions

            def calc(
                a: np.ndarray, b: np.ndarray, out: np.ndarray = None
            ) -> np.ndarray:
                if a.shape == b.shape:
                    # dot product between tensor and tensor
                    return np.einsum("ij...,jk...->ik...", a, b, out=out)  # type: ignore
                elif a.shape[1:] == b.shape:
                    # dot product between tensor and vector
                    return np.einsum("ij...,j...->i...", a, b, out=out)  # type: ignore
                else:
                    raise ValueError(f"Unsupported shapes ({a.shape}, {b.shape})")

            if conjugate:
                # create inner function calculating the dot product using conjugate

                def dot(
                    a: np.ndarray, b: np.ndarray, out: np.ndarray = None
                ) -> np.ndarray:
                    return calc(a, b.conjugate(), out=out)  # type: ignore

            else:
                dot = calc

        else:
            raise ValueError(f"Undefined backend `{backend}")

        return dot

    def get_dot_operator(self) -> Callable:
        """return operator calculating the dot product involving vector fields

        This supports both products between two vectors as well as products
        between a vector and a tensor.

        Warning:
            This function does not check types or dimensions.

        Returns:
            function that takes two instance of :class:`~numpy.ndarray`, which
            contain the discretized data of the two operands. An optional third
            argument can specify the output array to which the result is
            written. Note that the returned function is jitted with numba for
            speed.
        """
        # Deprecated this method on 2020-10-07
        import warnings

        warnings.warn(
            "get_dot_operator() method is deprecated. Use the `make_dot_operator()` "
            "method instead",
            DeprecationWarning,
        )
        return self.make_dot_operator()

    @fill_in_docstring
    def divergence(
        self,
        bc: "BoundariesData",
        out: Optional[VectorField] = None,
        *,
        label: str = "divergence",
    ) -> VectorField:
        """apply (tensor) divergence and return result as a field

        Args:
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            out (VectorField, optional):
                Optional scalar field to which the  result is written.
            label (str, optional):
                Name of the returned field

        Returns:
            VectorField: the result of applying the operator
        """
        tensor_divergence = self.grid.get_operator("tensor_divergence", bc=bc)
        if out is None:
            out = VectorField(self.grid, tensor_divergence(self.data), label=label)
        else:
            assert isinstance(out, VectorField), "Output must be a vector field"
            tensor_divergence(self.data, out=out.data)
        return out

    @property
    def integral(self):
        """ :class:`~numpy.ndarray`: integral of each component over space """
        return self.grid.integrate(self.data)

    def transpose(self, label: str = "transpose") -> "Tensor2Field":
        """return the transpose of the tensor field

        Args:
            label (str, optional): Name of the returned field

        Returns:
            Tensor2Field: holding the transpose of the tensor field
        """
        axes = (1, 0) + tuple(range(2, 2 + self.grid.dim))
        return Tensor2Field(self.grid, self.data.transpose(axes), label=label)

    def symmetrize(
        self, make_traceless: bool = False, inplace: bool = False
    ) -> "Tensor2Field":
        """symmetrize the tensor field in place

        Args:
            make_traceless (bool):
                Determines whether the result is also traceless
            inplace (bool):
                Flag determining whether to symmetrize the current field or
                return a new one
        """
        if inplace:
            out = self
        else:
            out = self.copy()

        out += self.transpose()
        out *= 0.5

        if make_traceless:
            dim = self.grid.dim
            value = self.trace() / dim
            for i in range(dim):
                out._data[i, i] -= value.data
        return out

    def to_scalar(
        self, scalar: str = "auto", *, label: Optional[str] = "scalar `{scalar}`"
    ) -> ScalarField:
        r""" return a scalar field by applying `method`
        
        The invariants of the tensor field :math:`\boldsymbol{A}` are
        
        .. math::
            I_1 &= \mathrm{tr}(\boldsymbol{A}) \\
            I_2 &= \frac12 \left[
                (\mathrm{tr}(\boldsymbol{A})^2 -
                \mathrm{tr}(\boldsymbol{A}^2)
            \right] \\
            I_3 &= \det(A)
            
        where `tr` denotes the trace and `det` denotes the determinant.
        Note that the three invariants can only be distinct and non-zero in three
        dimensions. In two dimensional spaces, we have the identity :math:`2 I_2 = I_3`
        and in one-dimensional spaces, we have :math:`I_1 = I_3` as well as
        :math:`I_2 = 0`.
            
        Args:
            scalar (str):
                The method to calculate the scalar. Possible choices include `norm` (the
                default), `min`, `max`, `squared_sum`, `norm_squared`, `trace` (or
                `invariant1`), `invariant2`, and `determinant` (or `invariant3`)
            label (str, optional):
                Name of the returned field
            
        Returns:
            :class:`pde.fields.scalar.ScalarField`: the scalar field after
            applying the operation
        """
        if scalar == "auto":
            scalar = "norm"

        if scalar == "norm":
            data = np.linalg.norm(self.data, axis=(0, 1))

        elif scalar == "min":
            data = np.min(self.data, axis=(0, 1))

        elif scalar == "max":
            data = np.max(self.data, axis=(0, 1))

        elif scalar == "squared_sum":
            data = np.sum(self.data ** 2, axis=(0, 1))

        elif scalar == "norm_squared":
            data = np.sum(self.data * self.data.conjugate(), axis=(0, 1))

        elif scalar == "trace" or scalar == "invariant1":
            data = self.data.trace(axis1=0, axis2=1)

        elif scalar == "invariant2":
            data = np.zeros(self.grid.shape)
            for i in range(self.grid.dim):
                for j in range(i):
                    data += (
                        self.data[i, i] * self.data[j, j]
                        - self.data[i, j] * self.data[j, i]
                    )
            data *= 0.5

        elif scalar in {"det", "determinant", "invariant3"}:
            if self.grid.dim == 1:
                data = self.data[0, 0]
            else:
                data = np.zeros(self.grid.shape)
                # this iterates over all of space and might thus be slow, but
                # the interface of np.linalg.det is not very flexible. We could
                # in principle use the definition of np.linalg.det without the
                # multiple checks to gain some speed
                for i in np.ndindex(self.grid.shape):
                    data[i] = np.linalg.det(self.data[(...,) + i])  # type: ignore

        else:
            raise ValueError(
                f"Unknown method `{scalar}` for `to_scalar`. Valid methods are `norm`, "
                "`min`, `max`, squared_sum`, `norm_squared`, `trace`, `determinant`, "
                "and `invariant#`, where # is 1, 2, or 3"
            )

        # determine label of the result
        if self.label is None:
            if label is not None:
                label = label.format(scalar=scalar)
        else:
            label = f"{scalar} of {self.label}"

        return ScalarField(self.grid, data, label=label)

    def trace(self, label: Optional[str] = "trace") -> ScalarField:
        """return the trace of the tensor field as a scalar field

        Args:
            label (str, optional): Name of the returned field

        Returns:
            ScalarField: holding the trace
        """
        return self.to_scalar(scalar="trace", label=label)



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np
import pytest

from pde.fields.base import FieldBase
from pde.fields.scalar import ScalarField
from pde.grids import CartesianGrid, PolarGrid, UnitGrid, boundaries
from pde.grids.tests.test_cartesian_grids import _get_cartesian_grid
from pde.tools.misc import module_available, skipUnlessModule


def test_interpolation_singular():
    """ test interpolation on singular dimensions """
    grid = UnitGrid([1])
    field = ScalarField(grid, data=3)

    # test constant boundary conditions
    bc = [{"type": "value", "value": 1}, {"type": "value", "value": 5}]
    x = np.linspace(0, 1, 7).reshape((7, 1))
    y = field.interpolate(x, backend="numba", bc=bc)
    np.testing.assert_allclose(y, 1 + 4 * x.ravel())

    # test derivative boundary conditions
    bc = [{"type": "derivative", "value": -2}, {"type": "derivative", "value": 2}]
    x = np.linspace(0, 1, 7).reshape((7, 1))
    y = field.interpolate(x, backend="numba", bc=bc)
    np.testing.assert_allclose(y, 2 + 2 * x.ravel())

    # test boundary interpolation
    for upper in [True, False]:
        val = field.get_boundary_values(axis=0, upper=upper, bc=[{"value": 1}])
        assert val == pytest.approx(1)


def test_simple_shapes(example_grid):
    """ test simple scalar fields """
    pf = ScalarField.random_uniform(example_grid)
    np.testing.assert_equal(pf.data.shape, example_grid.shape)
    pf_lap = pf.laplace("natural")
    np.testing.assert_equal(pf_lap.data.shape, example_grid.shape)
    assert isinstance(pf.integral, float)

    pf_c = pf.copy()
    np.testing.assert_allclose(pf.data, pf_c.data)
    assert pf.grid == pf_c.grid
    assert pf.data is not pf_c.data

    if module_available("matplotlib"):
        pf.plot()  # simply test whether this does not cause errors


def test_scalars():
    """ test some scalar fields """
    grid = CartesianGrid([[0.1, 0.3], [-2, 3]], [3, 4])
    s1 = ScalarField(grid, np.full(grid.shape, 1))
    s2 = ScalarField(grid, np.full(grid.shape, 2))
    assert s1.average == pytest.approx(1)
    assert s1.magnitude == pytest.approx(1)

    s3 = s1 + s2
    assert s3.grid == grid
    np.testing.assert_allclose(s3.data, 3)
    s1 += s2
    np.testing.assert_allclose(s1.data, 3)

    s2 = FieldBase.from_state(s1.attributes, data=s1.data)
    assert s1 == s2
    assert s1.grid is s2.grid

    attrs = ScalarField.unserialize_attributes(s1.attributes_serialized)
    s2 = FieldBase.from_state(attrs, data=s1.data)
    assert s1 == s2
    assert s1.grid is not s2.grid

    # test options for plotting images
    if module_available("matplotlib"):
        s1.plot(transpose=True, colorbar=True)

    s3 = ScalarField(grid, s1)
    assert s1 is not s3
    assert s1 == s3
    assert s1.grid is s3.grid

    # multiplication with numpy arrays
    arr = np.random.randn(*grid.shape)
    np.testing.assert_allclose((arr * s1).data, (s1 * arr).data)


def test_laplacian():
    """ test the gradient operator """
    grid = CartesianGrid([[0, 2 * np.pi], [0, 2 * np.pi]], [16, 16], periodic=True)
    s = ScalarField.random_harmonic(grid, axis_combination=np.add, modes=1)

    s_lap = s.laplace("natural")
    assert s_lap.data.shape == (16, 16)
    np.testing.assert_allclose(s_lap.data, -s.data, rtol=0.1, atol=0.1)

    s.laplace("natural", out=s_lap)
    assert s_lap.data.shape == (16, 16)
    np.testing.assert_allclose(s_lap.data, -s.data, rtol=0.1, atol=0.1)


def test_gradient():
    """ test the gradient operator """
    grid = CartesianGrid([[0, 2 * np.pi], [0, 2 * np.pi]], [16, 16], periodic=True)
    x, y = grid.cell_coords[..., 0], grid.cell_coords[..., 1]
    data = np.cos(x) + np.sin(y)

    s = ScalarField(grid, data)
    v = s.gradient("natural")
    assert v.data.shape == (2, 16, 16)
    np.testing.assert_allclose(v.data[0], -np.sin(x), rtol=0.1, atol=0.1)
    np.testing.assert_allclose(v.data[1], np.cos(y), rtol=0.1, atol=0.1)

    s.gradient("natural", out=v)
    assert v.data.shape == (2, 16, 16)
    np.testing.assert_allclose(v.data[0], -np.sin(x), rtol=0.1, atol=0.1)
    np.testing.assert_allclose(v.data[1], np.cos(y), rtol=0.1, atol=0.1)


def test_interpolation_to_grid(example_grid):
    """ test whether data is interpolated correctly for different grids """
    sf = ScalarField.random_uniform(example_grid)
    sf2 = sf.interpolate_to_grid(example_grid, backend="numba")
    np.testing.assert_allclose(sf.data, sf2.data, rtol=1e-6)


def test_insert_scalar(example_grid):
    """ test the `insert` method """
    f = ScalarField(example_grid)
    a = np.random.random()

    c = tuple(example_grid.point_to_cell(example_grid.get_random_point()))
    p = example_grid.cell_to_point(c, cartesian=False)
    f.insert(p, a)
    assert f.data[c] == pytest.approx(a / example_grid.cell_volumes[c])

    f.insert(example_grid.get_random_point(cartesian=False), a)
    assert f.integral == pytest.approx(2 * a)

    f.data = 0  # reset
    insert = example_grid.make_inserter_compiled()
    c = tuple(example_grid.point_to_cell(example_grid.get_random_point()))
    p = example_grid.cell_to_point(c, cartesian=False)
    insert(f.data, p, a)
    assert f.data[c] == pytest.approx(a / example_grid.cell_volumes[c])

    insert(f.data, example_grid.get_random_point(cartesian=False), a)
    assert f.integral == pytest.approx(2 * a)


def test_insert_1d():
    """ test the `insert` method for 1d systems """
    grid = PolarGrid(3, 5)
    f = ScalarField(grid)
    g = f.copy()
    a = np.random.random()
    for r in np.linspace(0, 3, 8).reshape(8, 1):
        f.data = g.data = 0
        f.insert(r, a)
        assert f.integral == pytest.approx(a)
        grid.make_inserter_compiled()(g.data, r, a)
        np.testing.assert_array_almost_equal(f.data, g.data)


def test_random_harmonic():
    """ test whether random harmonic fields behave correctly """
    grid = _get_cartesian_grid(2)  # get random Cartesian grid
    x = ScalarField.random_harmonic(grid, modes=1)
    scaling = sum((2 * np.pi / L) ** 2 for L in grid.cuboid.size)
    y = -x.laplace("natural") / scaling
    np.testing.assert_allclose(x.data, y.data, rtol=1e-2, atol=1e-2)


def test_get_line_data():
    """ test different extraction methods for line data """
    grid = UnitGrid([16, 32])
    c = ScalarField.random_harmonic(grid)

    np.testing.assert_equal(
        c.get_line_data(extract="cut_0"), c.get_line_data(extract="cut_x")
    )
    np.testing.assert_equal(
        c.get_line_data(extract="cut_1"), c.get_line_data(extract="cut_y")
    )

    for extract in ["project_0", "project_1", "project_x", "project_y"]:
        data = c.get_line_data(extract=extract)["data_y"]
        np.testing.assert_allclose(data, 0, atol=1e-14)


def test_from_expression():
    """ test creating scalar field from expression """
    grid = UnitGrid([1, 2])
    sf = ScalarField.from_expression(grid, "x * y", label="abc")
    assert sf.label == "abc"
    np.testing.assert_allclose(sf.data, [[0.25, 0.75]])


def test_interpolation_inhomogeneous_bc():
    """ test field interpolation with inhomogeneous boundary condition """
    sf = ScalarField(UnitGrid([3, 3], periodic=False))
    x = 1 + np.random.random()
    bc = ["natural", {"type": "value", "value": "x"}]
    v = sf.interpolate([x, 0], backend="numba", bc=bc)
    assert x == pytest.approx(v)


@skipUnlessModule("matplotlib")
def test_from_image(tmp_path):
    from matplotlib.pyplot import imsave

    img_data = np.random.uniform(size=(9, 8, 3))
    img_data_gray = img_data @ np.array([0.299, 0.587, 0.114])
    path = tmp_path / "test_from_image.png"
    imsave(path, img_data, vmin=0, vmax=1)
    sf = ScalarField.from_image(path)
    np.testing.assert_allclose(sf.data, img_data_gray.T[:, ::-1], atol=0.05)


def test_to_scalar():
    """ test conversion to scalar field """
    sf = ScalarField.random_uniform(UnitGrid([3, 3]))
    np.testing.assert_allclose(sf.to_scalar().data, sf.data)
    np.testing.assert_allclose(sf.to_scalar("norm_squared").data, sf.data ** 2)
    np.testing.assert_allclose(sf.to_scalar(lambda x: 2 * x).data, 2 * sf.data)

    with pytest.raises(ValueError):
        sf.to_scalar("nonsense")


@pytest.mark.parametrize("method", ["integral", "average"])
def test_projection(example_grid_nd, method):
    """ test scalar projection """
    sf = ScalarField.random_uniform(example_grid_nd)
    for ax in example_grid_nd.axes:
        sp = sf.project(ax, method=method)
        assert sp.grid.dim < example_grid_nd.dim
        assert sp.grid.num_axes == example_grid_nd.num_axes - 1
        if method == "integral":
            assert sp.integral == pytest.approx(sf.integral)
        elif method == "average":
            assert sp.average == pytest.approx(sf.average)

    with pytest.raises(ValueError):
        sf.project(example_grid_nd.axes[0], method="nonsense")


def test_slice(example_grid_nd):
    """ test scalar slicing """
    sf = ScalarField(example_grid_nd, 0.5)
    p = example_grid_nd.get_random_point()
    for i in range(example_grid_nd.num_axes):
        sf_slc = sf.slice({example_grid_nd.axes[i]: p[i]})
        np.testing.assert_allclose(sf_slc.data, 0.5)
        assert sf_slc.grid.dim < example_grid_nd.dim
        assert sf_slc.grid.num_axes == example_grid_nd.num_axes - 1

    with pytest.raises(boundaries.DomainError):
        sf.slice({example_grid_nd.axes[0]: -10})
    with pytest.raises(ValueError):
        sf.slice({"q": 0})


def test_slice_positions():
    """ test scalar slicing at standard positions """
    grid = UnitGrid([3, 1])
    sf = ScalarField(grid, np.arange(3).reshape(3, 1))
    assert sf.slice({"x": "min"}).data == 0
    assert sf.slice({"x": "mid"}).data == 1
    assert sf.slice({"x": "max"}).data == 2

    with pytest.raises(ValueError):
        sf.slice({"x": "foo"})
    with pytest.raises(ValueError):
        sf.slice({"x": 0}, method="nonsense")


def test_interpolation_mutable():
    """ test interpolation on mutable fields """
    grid = UnitGrid([2], periodic=True)
    field = ScalarField(grid)

    for backend in ["numba", "scipy"]:
        field.data = 1
        np.testing.assert_allclose(field.interpolate([0.5], backend=backend), 1)
        field.data = 2
        np.testing.assert_allclose(field.interpolate([0.5], backend=backend), 2)

    # test overwriting field values
    data = np.full_like(field.data, 3)
    intp = field.make_interpolator(backend="numba")
    np.testing.assert_allclose(intp(np.array([0.5]), data), 3)


def test_boundary_interpolation_1d():
    """ test boundary interpolation for 1d fields """
    grid = UnitGrid([5])
    field = ScalarField(grid, np.arange(grid.shape[0]))

    # test boundary interpolation
    bndry_val = 0.25
    for bndry in grid._iter_boundaries():
        val = field.get_boundary_values(*bndry, bc={"value": bndry_val})
        np.testing.assert_allclose(val, bndry_val)

        ev = field.make_get_boundary_values(*bndry, bc={"value": bndry_val})
        out = ev()
        np.testing.assert_allclose(out, bndry_val)
        ev(data=field.data, out=out)
        np.testing.assert_allclose(out, bndry_val)


def test_boundary_interpolation_2d():
    """ test boundary interpolation for 2d fields """
    grid = CartesianGrid([[0.1, 0.3], [-2, 3]], [3, 3])
    field = ScalarField.random_normal(grid)

    # test boundary interpolation
    bndry_val = np.random.randn(3)
    for bndry in grid._iter_boundaries():
        val = field.get_boundary_values(*bndry, bc={"value": bndry_val})
        np.testing.assert_allclose(val, bndry_val)

        ev = field.make_get_boundary_values(*bndry, bc={"value": bndry_val})
        out = ev()
        np.testing.assert_allclose(out, bndry_val)
        ev(data=field.data, out=out)
        np.testing.assert_allclose(out, bndry_val)


def test_numpy_ufuncs():
    """ test numpy ufuncs """
    grid = UnitGrid([2, 2])
    f1 = ScalarField.random_uniform(grid, 0.1, 0.9)

    f2 = np.sin(f1)
    np.testing.assert_allclose(f2.data, np.sin(f1.data))
    f2.data = 0
    np.sin(f1, out=f2)
    np.testing.assert_allclose(f2.data, np.sin(f1.data))

    np.testing.assert_allclose(np.add(f1, 2).data, f1.data + 2)

    with pytest.raises(TypeError):
        np.sum(f1, 1)


def test_plotting_1d():
    """ test plotting of 1d scalar fields """
    grid = UnitGrid([3])
    field = ScalarField.random_uniform(grid, 0.1, 0.9)

    ref = field.plot()
    field._update_plot(ref)


def test_plotting_2d():
    """ test plotting of 2d scalar fields """
    grid = UnitGrid([3, 3])
    field = ScalarField.random_uniform(grid, 0.1, 0.9)

    ref = field.plot()
    field._update_plot(ref)


@skipUnlessModule("napari")
def test_interactive_plotting():
    """ test the interactive plotting """
    grid = UnitGrid([3, 3])
    field = ScalarField.random_uniform(grid, 0.1, 0.9)
    field.plot_interactive(viewer_args={"show": False, "close": True})

    grid = UnitGrid([3, 3, 3])
    field = ScalarField.random_uniform(grid, 0.1, 0.9)
    field.plot_interactive(viewer_args={"show": False, "close": True})


def test_complex_dtype():
    """ test the support of a complex data type """
    grid = UnitGrid([2])
    f = ScalarField(grid, 1j)
    assert f.is_complex
    np.testing.assert_allclose(f.data, np.array([1j, 1j]))

    f = ScalarField(grid, 1)
    assert not f.is_complex
    with pytest.raises(np.core._exceptions.UFuncTypeError):
        f += 1j

    f = f + 1j
    assert f.is_complex
    np.testing.assert_allclose(f.data, np.full((2,), 1 + 1j))
    assert f.integral == pytest.approx(2 + 2j)
    assert f.average == pytest.approx(1 + 1j)
    np.testing.assert_allclose(f.to_scalar("abs").data, np.full((2,), np.sqrt(2)))
    assert f.magnitude == pytest.approx(np.sqrt(2))


def test_complex_plotting():
    """ test plotting of complex fields """
    for dim in (1, 2):
        f = ScalarField(UnitGrid([3] * dim), 1j)
        f.plot()


def test_complex_methods():
    """ test special methods for complex data type """
    grid = UnitGrid([2, 2])
    f = ScalarField(grid, 1j)
    for backend in ["scipy", "numba"]:
        val = f.interpolate([1, 1], backend=backend)
        np.testing.assert_allclose(val, np.array([1j, 1j]))

    f = ScalarField(grid, 1 + 2j)
    np.testing.assert_allclose(f.project("x").data, np.full((2,), 2 + 4j))
    np.testing.assert_allclose(f.slice({"x": 1}).data, np.full((2,), 1 + 2j))


def test_complex_operators():
    """ test differential operators for complex data type """
    f = ScalarField(UnitGrid([2, 2]), 1j)
    assert f.laplace("natural").magnitude == pytest.approx(0)



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np
import pytest

from pde import CartesianGrid, PolarGrid, ScalarField, SphericalGrid
from pde.grids.boundaries.local import NeumannBC


@pytest.mark.parametrize("grid_class", [PolarGrid, SphericalGrid])
def test_spherical_base_bcs(grid_class):
    """ test setting boundary conditions on spherical grids """
    grid = grid_class(2, 3)

    domain1 = grid.get_boundary_conditions(["derivative", {"type": "value"}])
    domain2 = grid.get_boundary_conditions({"type": "value"})
    assert domain1 == domain2

    # test boundary conditions for simulations with holes
    grid = grid_class((1, 2), 3)
    grid.get_boundary_conditions(["derivative", {"type": "value"}])
    domain1 = grid.get_boundary_conditions({"type": "value"})
    domain2 = grid.get_boundary_conditions(["value", "value"])
    assert domain1 == domain2


def test_polar_grid():
    """ test simple polar grid """
    grid = PolarGrid(4, 8)
    assert grid.dim == 2
    assert grid.numba_type == "f8[:]"
    assert grid.shape == (8,)
    assert not grid.has_hole
    assert grid.discretization[0] == pytest.approx(0.5)
    assert not grid.uniform_cell_volumes
    np.testing.assert_array_equal(grid.discretization, np.array([0.5]))
    assert grid.volume == pytest.approx(np.pi * 4 ** 2)
    assert grid.volume == pytest.approx(grid.integrate(1))

    np.testing.assert_allclose(grid.axes_coords[0], np.linspace(0.25, 3.75, 8))

    a = grid.get_operator("laplace", "natural")(np.random.random(8))
    assert a.shape == (8,)
    assert np.all(np.isfinite(a))

    # random points
    c = np.random.randint(8, size=(6, 1))
    p = grid.cell_to_point(c)
    np.testing.assert_array_equal(c, grid.point_to_cell(p))

    assert grid.contains_point(grid.get_random_point())
    assert grid.contains_point(grid.get_random_point(3.99))
    assert "laplace" in grid.operators


def test_polar_annulus():
    """ test simple polar grid with a hole """
    grid = PolarGrid((2, 4), 8)
    assert grid.dim == 2
    assert grid.numba_type == "f8[:]"
    assert grid.shape == (8,)
    assert grid.has_hole
    assert grid.discretization[0] == pytest.approx(0.25)
    assert not grid.uniform_cell_volumes
    np.testing.assert_array_equal(grid.discretization, np.array([0.25]))
    assert grid.volume == pytest.approx(np.pi * (4 ** 2 - 2 ** 2))
    assert grid.volume == pytest.approx(grid.integrate(1))
    assert grid.radius == (2, 4)

    np.testing.assert_allclose(grid.axes_coords[0], np.linspace(2.125, 3.875, 8))

    a = grid.get_operator("laplace", "natural")(np.random.random(8))
    assert a.shape == (8,)
    assert np.all(np.isfinite(a))

    # random points
    c = np.random.randint(8, size=(6, 1))
    p = grid.cell_to_point(c)
    np.testing.assert_array_equal(c, grid.point_to_cell(p))

    assert grid.contains_point(grid.get_random_point())
    assert grid.contains_point(grid.get_random_point(1.99))

    # test boundary points
    np.testing.assert_equal(grid._boundary_coordinates(0, False), np.array([2]))
    np.testing.assert_equal(grid._boundary_coordinates(0, True), np.array([4]))


def test_polar_to_cartesian():
    """ test conversion of polar grid to Cartesian """
    expr_pol = "1 / (1 + r**2)"
    expr_cart = expr_pol.replace("r**2", "(x**2 + y**2)")

    grid_pol = PolarGrid(7, 16)
    pf_pol = ScalarField.from_expression(grid_pol, expression=expr_pol)

    grid_cart = CartesianGrid([[-4, 4], [-3.9, 4.1]], [16, 16])
    pf_cart1 = pf_pol.interpolate_to_grid(grid_cart)
    pf_cart2 = ScalarField.from_expression(grid_cart, expression=expr_cart)
    np.testing.assert_allclose(pf_cart1.data, pf_cart2.data, atol=0.1)


def test_spherical_grid():
    """ test simple spherical grid """
    grid = SphericalGrid(4, 8)
    assert grid.dim == 3
    assert grid.numba_type == "f8[:]"
    assert grid.shape == (8,)
    assert not grid.has_hole
    assert grid.discretization[0] == pytest.approx(0.5)
    assert not grid.uniform_cell_volumes
    np.testing.assert_array_equal(grid.discretization, np.array([0.5]))
    assert grid.volume == pytest.approx(4 / 3 * np.pi * 4 ** 3)
    assert grid.volume == pytest.approx(grid.integrate(1))

    np.testing.assert_allclose(grid.axes_coords[0], np.linspace(0.25, 3.75, 8))

    a = grid.get_operator("laplace", "natural")(np.random.random(8))
    assert a.shape == (8,)
    assert np.all(np.isfinite(a))

    # random points
    c = np.random.randint(8, size=(6, 1))
    p = grid.cell_to_point(c)
    np.testing.assert_array_equal(c, grid.point_to_cell(p))

    assert grid.contains_point(grid.get_random_point())
    assert grid.contains_point(grid.get_random_point(3.99))
    assert "laplace" in grid.operators


def test_spherical_annulus():
    """ test simple spherical grid with a hole """
    grid = SphericalGrid((2, 4), 8)
    assert grid.dim == 3
    assert grid.numba_type == "f8[:]"
    assert grid.shape == (8,)
    assert grid.has_hole
    assert grid.discretization[0] == pytest.approx(0.25)
    assert not grid.uniform_cell_volumes
    np.testing.assert_array_equal(grid.discretization, np.array([0.25]))
    assert grid.volume == pytest.approx(4 / 3 * np.pi * (4 ** 3 - 2 ** 3))
    assert grid.volume == pytest.approx(grid.integrate(1))
    assert grid.radius == (2, 4)

    np.testing.assert_allclose(grid.axes_coords[0], np.linspace(2.125, 3.875, 8))

    a = grid.get_operator("laplace", "natural")(np.random.random(8))
    assert a.shape == (8,)
    assert np.all(np.isfinite(a))

    # random points
    c = np.random.randint(8, size=(6, 1))
    p = grid.cell_to_point(c)
    assert all(grid.contains_point(r) for r in p)
    np.testing.assert_array_equal(c, grid.point_to_cell(p))

    assert grid.contains_point(grid.get_random_point())
    assert grid.contains_point(grid.get_random_point(1.99))

    # test boundary points
    np.testing.assert_equal(grid._boundary_coordinates(0, False), np.array([2]))
    np.testing.assert_equal(grid._boundary_coordinates(0, True), np.array([4]))


def test_spherical_to_cartesian():
    """ test conversion of spherical grid to cartesian """
    expr_sph = "1 / (1 + r**2)"
    expr_cart = expr_sph.replace("r**2", "(x**2 + y**2 + z**2)")

    grid_sph = SphericalGrid(7, 16)
    pf_sph = ScalarField.from_expression(grid_sph, expression=expr_sph)

    grid_cart = CartesianGrid([[-4, 4], [-3.9, 4.1], [-4.1, 3.9]], [16] * 3)
    pf_cart1 = pf_sph.interpolate_to_grid(grid_cart)
    pf_cart2 = ScalarField.from_expression(grid_cart, expression=expr_cart)
    np.testing.assert_allclose(pf_cart1.data, pf_cart2.data, atol=0.1)


@pytest.mark.parametrize("grid_class", [PolarGrid, SphericalGrid])
def test_setting_boundary_conditions(grid_class):
    """ test setting some boundary conditions """
    grid = grid_class([0, 1], 3)
    b_inner = NeumannBC(grid, 0, upper=False)

    assert grid.get_boundary_conditions("natural")[0].low == b_inner
    assert grid.get_boundary_conditions({"value": 2})[0].low == b_inner
    bcs = grid.get_boundary_conditions(["value", "value"])
    assert bcs[0].low != b_inner

    grid = grid_class([1, 2], 3)
    bcs = grid.get_boundary_conditions(["value", "value"])
    assert bcs[0].low != b_inner



"""
Module defining classes for time intervals for trackers

The provided interval classes are:

.. autosummary::
   :nosignatures:

   ConstantIntervals
   LogarithmicIntervals
   RealtimeIntervals
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

import copy
import math
import time
from typing import Any, Dict, Optional, Union

from ..tools.parse_duration import parse_duration

InfoDict = Optional[Dict[str, Any]]


class ConstantIntervals:
    """ class representing equidistantly spaced time intervals """

    def __init__(self, dt: float = 1, t_start: Optional[float] = None):
        """
        Args:
            dt (float): The duration between subsequent intervals. This is
                measured in simulation time units.
            t_start (float, optional): The time after which the tracker becomes
                active. If omitted, the tracker starts recording right away.
                This argument can be used for an initial equilibration period
                during which no data is recorded.
        """
        self.dt = float(dt)
        self.t_start = t_start
        self._t_next: Optional[float] = None  # next time it should be called

    def __repr__(self):
        return f"{self.__class__.__name__}(dt={self.dt:g}, t_start={self.t_start})"

    def copy(self):
        """ return a copy of this instance """
        return copy.copy(self)

    def _initialize(self, t: float) -> float:
        """initialize the tracker

        Args:
            t (float): The starting time of the simulation

        Returns:
            float: The first time the tracker needs to handle data
        """
        if self.t_start is None:
            self._t_next = t
        else:
            self._t_next = max(t, self.t_start)
        return self._t_next

    def next(self, t: float) -> float:
        """computes the next time point based on the current time t

        Args:
            t (float): The current time point of the simulation
        """
        if self._t_next is None:
            self._initialize(t)

        self._t_next = self._t_next + self.dt  # type: ignore

        # make sure that the new time `_t_next` is larger than t
        while self._t_next <= t:
            self._t_next += self.dt
        return self._t_next


class LogarithmicIntervals(ConstantIntervals):
    """ class representing logarithmically spaced time intervals """

    def __init__(
        self, dt_initial: float = 1, factor: float = 1, t_start: Optional[float] = None
    ):
        """
        Args:
            dt_initial (float): The initial duration between subsequent
                intervals. This is measured in simulation time units.
            factor (float): The factor by which the time between intervals is
                increased every time. Values larger than one lead to time
                intervals that are increasingly further apart.
            t_start (float, optional): The time after which the tracker becomes
                active. If omitted, the tracker starts recording right away.
                This argument can be used for an initial equilibration period
                during which no data is recorded.
        """
        super().__init__(dt=dt_initial / factor, t_start=t_start)
        self.factor = float(factor)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(dt={self.dt:g}, "
            f"factor={self.factor:g}, t_start={self.t_start})"
        )

    def next(self, t: float) -> float:
        """computes the next time point based on the current time t

        Args:
            t (float): The current time point of the simulation
        """
        self.dt *= self.factor
        return super().next(t)


class RealtimeIntervals(ConstantIntervals):
    """class representing time intervals spaced equidistantly in real time

    This spacing is only achieved approximately and depends on the initial value
    set by `dt_initial` and the actual variation in computation speed.
    """

    def __init__(self, duration: Union[float, str], dt_initial: float = 0.01):
        """
        Args:
            duration (float or str): The duration (in realtime seconds) that the
                intervals should be spaced apart. The duration can also be given
                as a string, which is then parsed using the function
                :func:`~pde.tools.parse_duration.parse_duration`.
            dt_initial (float): The initial duration between subsequent
                intervals. This is measured in simulation time units.
        """
        super().__init__(dt=dt_initial)
        try:
            self.duration = float(duration)
        except Exception:
            td = parse_duration(str(duration))
            self.duration = td.total_seconds()
        self._last_time: Optional[float] = None

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(duration={self.duration:g}, "
            f"dt_initial={self.dt:g})"
        )

    def _initialize(self, t: float) -> float:
        """initialize the tracker

        Args:
            t (float): The starting time of the simulation

        Returns:
            float: The first time the tracker needs to handle data
        """
        self._last_time = time.time()
        return super()._initialize(t)

    def next(self, t: float) -> float:
        """computes the next time point based on the current time t

        Args:
            t (float): The current time point of the simulation
        """
        if self._last_time is None:
            self._last_time = time.time()
        else:
            # adapt time step
            current_time = time.time()
            time_passed = current_time - self._last_time
            if time_passed > 0:
                dt_predict = self.dt * self.duration / time_passed
                # use geometric average to provide some smoothing
                self.dt = math.sqrt(self.dt * dt_predict)
            else:
                self.dt *= 2
            self._last_time = current_time
        return super().next(t)


IntervalType = ConstantIntervals
IntervalData = Union[IntervalType, float, int, str]


def get_interval(interval: IntervalData) -> IntervalType:
    """create IntervalType from various data formats

    If interval is of type :class:`IntervalType` it is simply returned
    """
    if isinstance(interval, IntervalType):
        return interval
    elif isinstance(interval, (int, float)):
        return ConstantIntervals(interval)
    elif isinstance(interval, str):
        return RealtimeIntervals(interval)
    else:
        raise TypeError(f"Do not understand interval type {interval}")



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np
import pytest

from pde import PDE, DiffusionPDE, FieldCollection, ScalarField, UnitGrid
from pde.solvers import (
    Controller,
    ExplicitSolver,
    ImplicitSolver,
    ScipySolver,
    registered_solvers,
)


def test_solver_registration():
    """ test solver registration """
    solvers = registered_solvers()
    assert "explicit" in solvers
    assert "scipy" in solvers


def test_solver_in_pde_class():
    """ test whether solver instances can be used in pde instances """
    field = ScalarField.random_uniform(UnitGrid([16, 16]), -1, 1)
    eq = DiffusionPDE()
    eq.solve(field, t_range=1, method=ScipySolver, tracker=None)


@pytest.mark.parametrize("solver_class", [ExplicitSolver, ImplicitSolver, ScipySolver])
def test_compare_solvers(solver_class):
    """ compare several solvers """
    field = ScalarField.random_uniform(UnitGrid([8, 8]), -1, 1)
    eq = DiffusionPDE()

    # ground truth
    c1 = Controller(ExplicitSolver(eq, scheme="runge-kutta"), t_range=0.1, tracker=None)
    s1 = c1.run(field, dt=5e-3)

    c2 = Controller(solver_class(eq), t_range=0.1, tracker=None)
    with np.errstate(under="ignore"):
        s2 = c2.run(field, dt=5e-3)

    np.testing.assert_allclose(s1.data, s2.data, rtol=1e-2, atol=1e-2)


@pytest.mark.parametrize("backend", ["numpy", "numba"])
def test_solvers_complex(backend):
    """ test solvers with a complex PDE """
    r = FieldCollection.scalar_random_uniform(2, UnitGrid([3]), labels=["a", "b"])
    c = r["a"] + 1j * r["b"]
    assert c.is_complex

    # assume c = a + i * b
    eq_c = PDE({"c": "-I * laplace(c)"})
    eq_r = PDE({"a": "laplace(b)", "b": "-laplace(a)"})
    res_r = eq_r.solve(r, t_range=1e-2, dt=1e-3, backend="numpy", tracker=None)
    exp_c = res_r[0].data + 1j * res_r[1].data

    for solver_class in [ExplicitSolver, ImplicitSolver, ScipySolver]:
        solver = solver_class(eq_c, backend=backend)
        controller = Controller(solver, t_range=1e-2, tracker=None)
        res_c = controller.run(c, dt=1e-3)
        np.testing.assert_allclose(res_c.data, exp_c, rtol=1e-3, atol=1e-3)



#!/usr/bin/env python3
"""
This script tests the performance of the implementation of different boundary conditions
"""

import sys
from pathlib import Path

PACKAGE_PATH = Path(__file__).resolve().parents[1]
sys.path.append(str(PACKAGE_PATH))

import numpy as np

from pde import ScalarField, UnitGrid
from pde.grids.boundaries.axes import Boundaries
from pde.tools.misc import estimate_computation_speed


def main():
    """ main routine testing the performance """
    print("Reports calls-per-second (larger is better)\n")

    # Cartesian grid with different shapes and boundary conditions
    for size in [32, 512]:
        grid = UnitGrid((size, size), periodic=False)
        print(grid)

        field = ScalarField.random_normal(grid)
        bc_value = np.ones(size)
        result = field.laplace(bc={"value": 1}).data

        for bc in ["scalar", "array", "linked"]:
            if bc == "scalar":
                bcs = {"value": 1}
            elif bc == "array":
                bcs = {"value": bc_value}
            elif bc == "linked":
                bcs = Boundaries.from_data(grid, {"value": bc_value}, rank=0)
                for ax, upper in grid._iter_boundaries():
                    bcs[ax][upper].link_value(bc_value)
            # result = field.laplace(bc=bcs).data
            laplace = grid.get_operator("laplace", bc=bcs)
            # call once to pre-compile and test result
            np.testing.assert_allclose(laplace(field.data), result)
            speed = estimate_computation_speed(laplace, field.data)
            print(f"{bc:>6s}:{int(speed):>9d}")

        print()


if __name__ == "__main__":
    main()



"""
Defines the :class:`~pde.controller.Controller` class for solving pdes.

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""


import datetime
import logging
import time
from typing import TYPE_CHECKING, Any, Dict, Tuple, TypeVar, Union  # @UnusedImport

from ..trackers.base import (
    FinishedSimulation,
    TrackerCollection,
    TrackerCollectionDataType,
)
from ..version import __version__
from .base import SolverBase

if TYPE_CHECKING:
    from ..fields.base import FieldBase  # @UnusedImport


TRangeType = Union[float, Tuple[float, float]]
TState = TypeVar("TState", bound="FieldBase")


class Controller:
    """ class controlling a simulation """

    _t_range: Tuple[float, float]

    def __init__(
        self,
        solver: SolverBase,
        t_range: TRangeType,
        tracker: TrackerCollectionDataType = ["progress", "consistency"],
    ):
        """
        Args:
            solver (:class:`~pde.solvers.base.SolverBase`):
                Solver instance that is used to advance the simulation in time
            t_range (float or tuple):
                Sets the time range for which the simulation is run. If only a
                single value `t_end` is given, the time range is assumed to be
                `[0, t_end]`.
            tracker:
                Defines trackers that process the state of the simulation at
                fixed time intervals. Multiple trackers can be specified as a
                list. The default value 'auto' is converted to
                `['progress', 'consistency']` for normal simulations. This thus
                displays a progress bar and checks the state for consistency,
                aborting the simulation when not-a-number values appear. To
                disable trackers, set the value to `None`.
        """
        self.solver = solver
        self.t_range = t_range  # type: ignore
        self.trackers = TrackerCollection.from_data(tracker)

        self.info: Dict[str, Any] = {"package_version": __version__}
        self._logger = logging.getLogger(self.__class__.__name__)

    @property
    def t_range(self) -> Tuple[float, float]:
        return self._t_range

    @t_range.setter
    def t_range(self, value: TRangeType):
        # determine time range
        try:
            self._t_range = 0, float(value)  # type: ignore
        except TypeError:  # assume a single number was given
            if len(value) == 2:  # type: ignore
                self._t_range = tuple(value)  # type: ignore
            else:
                raise ValueError(
                    "t_range must be set to a single number or a tuple of two numbers"
                )

    def run(self, state: TState, dt: float = None) -> TState:
        """run the simulation

        Diagnostic information about the solver procedure are available in the
        `diagnostics` property of the instance after this function has been
        called.

        Args:
            state:
                The initial state of the simulation. This state will be copied
                and thus not modified by the simulation. Instead, the final
                state will be returned and trackers can be used to record
                intermediate states.
            dt (float):
                Time step of the chosen stepping scheme. If `None`, a default
                value based on the stepper will be chosen.

        Returns:
            The state at the final time point.
        """
        # copy the initial state to not modify the supplied one
        if hasattr(self.solver, "pde") and self.solver.pde.complex_valued:
            self._logger.info("Convert state to complex numbers")
            state = state.copy(dtype="complex")
        else:
            state = state.copy()
        t_start, t_end = self.t_range

        # initialize solver information
        self.info["t_start"] = t_start
        self.info["t_end"] = t_end
        self.info["solver_class"] = self.solver.__class__.__name__
        self.diagnostics: Dict[str, Any] = {
            "controller": self.info,
            "solver": self.solver.info,
        }

        # initialize trackers
        self.trackers.initialize(state, info=self.diagnostics)

        def _handle_stop_iteration(err):
            """ helper function for handling interrupts raised by trackers """
            if isinstance(err, FinishedSimulation):
                # tracker determined that the simulation finished
                self.info["successful"] = True
                msg = f"Simulation finished at t={t}"
                msg_level = logging.INFO
                if err.value:
                    self.info["stop_reason"] = err.value
                    msg += f" ({err.value})"
                else:
                    self.info["stop_reason"] = "Tracker raised FinishedSimulation"

            else:
                # tracker determined that there was a problem
                self.info["successful"] = False
                msg = f"Simulation aborted at t={t}"
                msg_level = logging.WARNING
                if err.value:
                    self.info["stop_reason"] = err.value
                    msg += f" ({err.value})"
                else:
                    self.info["stop_reason"] = "Tracker raised StopIteration"

            return msg_level, msg

        # initialize the stepper
        stepper = self.solver.make_stepper(state=state, dt=dt)

        # initialize profiling information
        solver_start = datetime.datetime.now()
        self.info["solver_start"] = str(solver_start)
        profiler = {"solver": 0.0, "tracker": 0.0}
        self.info["profiler"] = profiler
        prof_start_tracker = time.process_time()

        # add some tolerance to account for inaccurate float point math
        if dt is None:
            dt = self.solver.info.get("dt")
            # Note that self.solver.info['dt'] might be None

        if dt is None:
            atol = 1e-12
        else:
            atol = 0.1 * dt

        # evolve the system from t_start to t_end
        t = t_start
        self._logger.debug(f"Start simulation at t={t}")
        try:
            while t < t_end:
                # determine next time point with an action
                t_next_action = self.trackers.handle(state, t, atol=atol)
                t_next_action = max(t_next_action, t + atol)
                t_break = min(t_next_action, t_end)

                prof_start_solve = time.process_time()
                profiler["tracker"] += prof_start_solve - prof_start_tracker

                # advance the system to the new time point
                t = stepper(state, t, t_break)

                prof_start_tracker = time.process_time()
                profiler["solver"] += prof_start_tracker - prof_start_solve

        except StopIteration as err:
            # iteration has been interrupted by a tracker
            msg_level, msg = _handle_stop_iteration(err)

        except KeyboardInterrupt:
            # iteration has been interrupted by the user
            self.info["successful"] = False
            self.info["stop_reason"] = "User interrupted simulation"
            msg = f"Simulation interrupted at t={t}"
            msg_level = logging.INFO

        else:
            # reached final time
            self.info["successful"] = True
            self.info["stop_reason"] = "Reached final time"
            msg = f"Simulation finished at t={t_end}."
            msg_level = logging.INFO

            # handle trackers one more time when t_end is reached
            try:
                self.trackers.handle(state, t, atol=atol)
            except StopIteration as err:
                # error detected in the final handling of the tracker
                msg_level, msg = _handle_stop_iteration(err)

        # calculate final statistics
        profiler["tracker"] += time.process_time() - prof_start_tracker
        duration = datetime.datetime.now() - solver_start
        self.info["solver_duration"] = str(duration)
        self.info["t_final"] = t
        self.trackers.finalize(info=self.diagnostics)

        # show information after a potential progress bar has been deleted to
        # not mess up the display
        self._logger.log(msg_level, msg)
        if profiler["tracker"] > max(profiler["solver"], 1):
            self._logger.warning(
                f"Spent more time on handling trackers ({profiler['tracker']}) than on "
                f"the actual simulation ({profiler['solver']})"
            )

        return state



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from pde.tools.plotting import plot_on_axes, plot_on_figure


def test_plot_on_axes(tmp_path):
    """ test the plot_on_axes decorator """

    @plot_on_axes
    def plot(ax):
        ax.plot([0, 1], [0, 1])

    path = tmp_path / "test.png"
    plot(title="Test", filename=path)
    assert path.stat().st_size > 0


def test_plot_on_figure(tmp_path):
    """ test the plot_on_figure decorator """

    @plot_on_figure
    def plot(fig):
        ax1, ax2 = fig.subplots(1, 2)
        ax1.plot([0, 1], [0, 1])
        ax2.plot([0, 1], [0, 1])

    path = tmp_path / "test.png"
    plot(title="Test", filename=path)
    assert path.stat().st_size > 0



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from pde.tools import output


def test_progress_bars():
    """ test progress bars """
    for pb_cls in [output.SimpleProgress, output.get_progress_bar_class()]:
        tot = 0
        for i in pb_cls(range(4)):
            tot += i
        assert tot == 6


def test_in_jupyter_notebook():
    """ test the function in_jupyter_notebook """
    assert isinstance(output.in_jupyter_notebook(), bool)


def test_display_progress(capsys):
    """ test whether this works """
    for _ in output.display_progress(range(2)):
        pass
    out, err = capsys.readouterr()
    assert out == ""
    assert len(err) > 0



r"""
This module implements differential operators on spherical grids 

.. autosummary::
   :nosignatures:

   make_laplace
   make_gradient
   make_divergence
   make_vector_gradient
   make_tensor_divergence
   
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from typing import Tuple

import numpy as np

from ...tools.docstrings import fill_in_docstring
from ...tools.numba import jit_allocate_out
from ...tools.typing import OperatorType
from ..boundaries import Boundaries
from ..spherical import SphericalGrid
from .common import make_general_poisson_solver


@SphericalGrid.register_operator("laplace", rank_in=0, rank_out=0)
@fill_in_docstring
def make_laplace(bcs: Boundaries, conservative: bool = True) -> OperatorType:
    """make a discretized laplace operator for a spherical grid

    {DESCR_SPHERICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        conservative (bool): flag indicating whether the laplace operator should
            be conservative (which results in slightly slower computations).

    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(0)

    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    dr = bcs.grid.discretization[0]
    rs = bcs.grid.axes_coords[0]
    r_min, r_max = bcs.grid.axes_bounds[0]

    # prepare boundary values
    value_lower_bc = bcs[0].low.make_virtual_point_evaluator()
    value_upper_bc = bcs[0].high.make_virtual_point_evaluator()

    if conservative:
        # create a conservative spherical laplace operator
        rl = r_min + dr * np.arange(dim_r)  # inner radii of spherical shells
        rh = rl + dr  # outer radii
        assert np.isclose(rh[-1], r_max)
        volumes = (rh ** 3 - rl ** 3) / 3  # volume of the spherical shells
        factor_l = (rs - 0.5 * dr) ** 2 / (dr * volumes)
        factor_h = (rs + 0.5 * dr) ** 2 / (dr * volumes)

        @jit_allocate_out(out_shape=(dim_r,))
        def laplace(arr, out=None):
            """ apply laplace operator to array `arr` """
            i = 0
            out[i] = factor_h[i] * (arr[i + 1] - arr[i])
            if r_min > 0:
                arr_r_l = value_lower_bc(arr, (i,))
                out[i] -= factor_l[i] * (arr[i] - arr_r_l)

            for i in range(1, dim_r - 1):  # iterate inner radial points
                out[i] = factor_h[i] * (arr[i + 1] - arr[i])
                out[i] -= factor_l[i] * (arr[i] - arr[i - 1])

            # express boundary condition at outer side
            i = dim_r - 1
            arr_r_h = value_upper_bc(arr, (i,))
            out[i] = factor_h[i] * (arr_r_h - arr[i])
            out[i] -= factor_l[i] * (arr[i] - arr[i - 1])
            return out

    else:  # create an operator that is not conservative
        dr2 = 1 / dr ** 2

        @jit_allocate_out(out_shape=(dim_r,))
        def laplace(arr, out=None):
            """ apply laplace operator to array `arr` """
            i = 0
            if r_min == 0:
                out[i] = 3 * (arr[i + 1] - arr[i]) * dr2
            else:
                arr_r_l = value_lower_bc(arr, (i,))
                out[i] = (arr[i + 1] - 2 * arr[i] + arr_r_l) * dr2
                out[i] += (arr[i + 1] - arr_r_l) / (rs[i] * dr)

            for i in range(1, dim_r - 1):  # iterate inner radial points
                out[i] = (arr[i + 1] - 2 * arr[i] + arr[i - 1]) * dr2
                out[i] += (arr[i + 1] - arr[i - 1]) / (rs[i] * dr)

            # express boundary condition at outer side
            i = dim_r - 1
            arr_r_h = value_upper_bc(arr, (i,))
            out[i] = (arr_r_h - 2 * arr[i] + arr[i - 1]) * dr2
            out[i] += (arr_r_h - arr[i - 1]) / (rs[i] * dr)
            return out

    return laplace  # type: ignore


@SphericalGrid.register_operator("gradient", rank_in=0, rank_out=1)
@fill_in_docstring
def make_gradient(bcs: Boundaries) -> OperatorType:
    """make a discretized gradient operator for a spherical grid

    {DESCR_SPHERICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(0)

    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    dr = bcs.grid.discretization[0]
    r_min, _ = bcs.grid.axes_bounds[0]

    scale_r = 1 / (2 * dr)

    # prepare boundary values
    boundary = bcs[0]
    value_lower_bc = boundary.low.make_virtual_point_evaluator()
    value_upper_bc = boundary.high.make_virtual_point_evaluator()

    @jit_allocate_out(out_shape=(3, dim_r))
    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        i = 0
        if r_min == 0:
            # Apply Neumann condition at the origin
            out[0, i] = (arr[1] - arr[0]) * scale_r
        else:
            arr_r_l = value_lower_bc(arr, (i,))
            out[0, i] = (arr[1] - arr_r_l) * scale_r
        out[1, i] = out[2, i] = 0  # no angular dependence by definition

        for i in range(1, dim_r - 1):  # iterate inner radial points
            out[0, i] = (arr[i + 1] - arr[i - 1]) * scale_r
            out[1, i] = out[2, i] = 0  # no angular dependence by definition

        i = dim_r - 1
        arr_r_h = value_upper_bc(arr, (i,))
        out[0, i] = (arr_r_h - arr[i - 1]) * scale_r
        out[1, i] = out[2, i] = 0  # no angular dependence by definition

        return out

    return gradient  # type: ignore


@SphericalGrid.register_operator("gradient_squared", rank_in=0, rank_out=0)
@fill_in_docstring
def make_gradient_squared(bcs: Boundaries, central: bool = True) -> OperatorType:
    """make a discretized gradient squared operator for a spherical grid

    {DESCR_SPHERICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        central (bool):
            Whether a central difference approximation is used for the gradient
            operator. If this is False, the squared gradient is calculated as
            the mean of the squared values of the forward and backward
            derivatives.

    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(0)

    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    r_min, _ = bcs.grid.axes_bounds[0]
    dr = bcs.grid.discretization[0]

    # prepare boundary values
    boundary = bcs[0]
    value_lower_bc = boundary.low.make_virtual_point_evaluator()
    value_upper_bc = boundary.high.make_virtual_point_evaluator()

    if central:
        # use central differences
        scale = 1 / (2 * dr) ** 2

        @jit_allocate_out(out_shape=(dim_r,))
        def gradient_squared(arr, out=None):
            """ apply squared gradient operator to array `arr` """
            if r_min == 0:
                # Apply Neumann condition at the origin
                out[0] = (arr[1] - arr[0]) ** 2 * scale
            else:
                arr_r_l = value_lower_bc(arr, (0,))
                out[0] = (arr[1] - arr_r_l) ** 2 * scale

            for i in range(1, dim_r - 1):  # iterate inner radial points
                out[i] = (arr[i + 1] - arr[i - 1]) ** 2 * scale

            i = dim_r - 1
            arr_r_h = value_upper_bc(arr, (i,))
            out[i] = (arr_r_h - arr[i - 1]) ** 2 * scale

            return out

    else:
        # use forward and backward differences
        scale = 1 / (2 * dr ** 2)

        @jit_allocate_out(out_shape=(dim_r,))
        def gradient_squared(arr, out=None):
            """ apply squared gradient operator to array `arr` """
            if r_min == 0:
                # Apply Neumann condition at the origin
                out[0] = (arr[1] - arr[0]) ** 2 * scale
            else:
                arr_r_l = value_lower_bc(arr, (0,))
                out[0] = ((arr[1] - arr[0]) ** 2 + (arr[0] - arr_r_l) ** 2) * scale

            for i in range(1, dim_r - 1):  # iterate inner radial points
                term = (arr[i + 1] - arr[i]) ** 2 + (arr[i] - arr[i - 1]) ** 2
                out[i] = term * scale

            i = dim_r - 1
            arr_r_h = value_upper_bc(arr, (i,))
            out[i] = ((arr_r_h - arr[i]) ** 2 + (arr[i] - arr[i - 1]) ** 2) * scale

            return out

    return gradient_squared  # type: ignore


@SphericalGrid.register_operator("divergence", rank_in=1, rank_out=0)
@fill_in_docstring
def make_divergence(bcs: Boundaries) -> OperatorType:
    """make a discretized divergence operator for a spherical grid

    {DESCR_SPHERICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(0)

    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    dr = bcs.grid.discretization[0]
    rs = bcs.grid.axes_coords[0]
    r_min, _ = bcs.grid.axes_bounds[0]

    scale_r = 1 / (2 * dr)

    # prepare boundary values
    boundary = bcs[0]
    value_lower_bc = boundary.low.make_virtual_point_evaluator()
    value_upper_bc = boundary.high.make_virtual_point_evaluator()

    if r_min == 0:

        @jit_allocate_out(out_shape=(dim_r,))
        def divergence(arr, out=None):
            """ apply divergence operator to array `arr` """
            i = 0
            out[i] = (arr[0, 1] + 7 * arr[0, 0]) * scale_r

            for i in range(1, dim_r - 1):  # iterate inner radial points
                out[i] = (arr[0, i + 1] - arr[0, i - 1]) * scale_r
                out[i] += 2 * arr[0, i] / ((i + 0.5) * dr)

            i = dim_r - 1
            arr_r_h = value_upper_bc(arr[0], (i,))
            out[i] = (arr_r_h - arr[0, i - 1]) * scale_r
            out[i] += 2 * arr[0, i] / ((i + 0.5) * dr)

            return out

    else:  # r_min > 0
        fs = 2 / rs  # factors that need to be multiplied below

        @jit_allocate_out(out_shape=(dim_r,))
        def divergence(arr, out=None):
            """ apply divergence operator to array `arr` """
            # inner radial boundary condition
            i = 0
            arr_r_l = value_lower_bc(arr[0], (i,))
            out[i] = (arr[0, i + 1] - arr_r_l) * scale_r + fs[i] * arr[0, i]

            for i in range(1, dim_r - 1):  # iterate radial points
                out[i] = (arr[0, i + 1] - arr[0, i - 1]) * scale_r + fs[i] * arr[0, i]

            # outer radial boundary condition
            i = dim_r - 1
            arr_r_h = value_upper_bc(arr[0], (i,))
            out[i] = (arr_r_h - arr[0, i - 1]) * scale_r + fs[i] * arr[0, i]

            return out

    return divergence  # type: ignore


@SphericalGrid.register_operator("vector_gradient", rank_in=1, rank_out=2)
@fill_in_docstring
def make_vector_gradient(bcs: Boundaries) -> OperatorType:
    """make a discretized vector gradient operator for a spherical grid

    {DESCR_SPHERICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(1)

    gradient_r = make_gradient(bcs.extract_component(0))
    gradient_theta = make_gradient(bcs.extract_component(1))
    gradient_phi = make_gradient(bcs.extract_component(2))

    @jit_allocate_out(out_shape=(3, 3) + bcs.grid.shape)
    def vector_gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        gradient_r(arr[0], out=out[:, 0])
        gradient_theta(arr[1], out=out[:, 1])
        gradient_phi(arr[2], out=out[:, 2])
        return out

    return vector_gradient  # type: ignore


@SphericalGrid.register_operator("tensor_divergence", rank_in=2, rank_out=1)
@fill_in_docstring
def make_tensor_divergence(bcs: Boundaries) -> OperatorType:
    """make a discretized tensor divergence operator for a spherical grid

    {DESCR_SPHERICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(1)

    divergence_r = make_divergence(bcs.extract_component(0))
    divergence_theta = make_divergence(bcs.extract_component(1))
    divergence_phi = make_divergence(bcs.extract_component(2))

    @jit_allocate_out(out_shape=(3,) + bcs.grid.shape)
    def tensor_divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        divergence_r(arr[0], out=out[0])
        divergence_theta(arr[1], out=out[1])
        divergence_phi(arr[2], out=out[2])
        return out

    return tensor_divergence  # type: ignore


@fill_in_docstring
def _get_laplace_matrix(bcs: Boundaries) -> Tuple[np.ndarray, np.ndarray]:
    """get sparse matrix for laplace operator on a polar grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        tuple: A sparse matrix and a sparse vector that can be used to evaluate
        the discretized laplacian
    """
    from scipy import sparse

    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(0)

    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    dr = bcs.grid.discretization[0]
    rs = bcs.grid.axes_coords[0]
    r_min, r_max = bcs.grid.axes_bounds[0]

    # create a conservative spherical laplace operator
    rl = r_min + dr * np.arange(dim_r)  # inner radii of spherical shells
    rh = rl + dr  # outer radii
    assert np.isclose(rh[-1], r_max)
    volumes = (rh ** 3 - rl ** 3) / 3  # volume of the spherical shells

    factor_l = (rs - 0.5 * dr) ** 2 / (dr * volumes)
    factor_h = (rs + 0.5 * dr) ** 2 / (dr * volumes)

    matrix = sparse.dok_matrix((dim_r, dim_r))
    vector = sparse.dok_matrix((dim_r, 1))

    for i in range(dim_r):
        matrix[i, i] += -factor_l[i] - factor_h[i]

        if i == 0:
            if r_min == 0:
                matrix[i, i + 1] = factor_l[i]
            else:
                const, entries = bcs[0].get_data((-1,))
                vector[i] += const * factor_l[i]
                for k, v in entries.items():
                    matrix[i, k] += v * factor_l[i]

        else:
            matrix[i, i - 1] = factor_l[i]

        if i == dim_r - 1:
            const, entries = bcs[0].get_data((dim_r,))
            vector[i] += const * factor_h[i]
            for k, v in entries.items():
                matrix[i, k] += v * factor_h[i]

        else:
            matrix[i, i + 1] = factor_h[i]

    return matrix, vector


@SphericalGrid.register_operator("poisson_solver", rank_in=0, rank_out=0)
@fill_in_docstring
def make_poisson_solver(bcs: Boundaries, method: str = "auto") -> OperatorType:
    """make a operator that solves Poisson's equation

    {DESCR_POLAR_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    matrix, vector = _get_laplace_matrix(bcs)
    return make_general_poisson_solver(matrix, vector, method)



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import random

import numpy as np
import pytest
from scipy import ndimage

from pde import CartesianGrid, ScalarField, UnitGrid
from pde.grids.operators import cartesian as ops

π = np.pi


def _get_random_grid_bcs(ndim: int, dx="random", periodic="random", rank=0):
    """ create a random Cartesian grid with natural bcs """
    shape = np.random.randint(2, 5, ndim)

    if dx == "random":
        dx = np.random.uniform(0.5, 1.5, ndim)
    elif dx == "uniform":
        dx = np.full(ndim, np.random.uniform(0.5, 1.5))
    else:
        dx = np.broadcast_to(dx, shape)

    if periodic == "random":
        periodic = random.choice([True, False])

    sizes = [(0, float(s * d)) for s, d in zip(shape, dx)]
    grid = CartesianGrid(sizes, shape, periodic=periodic)
    return grid.get_boundary_conditions("natural", rank=rank)


@pytest.mark.parametrize("periodic", [True, False])
def test_singular_dimensions_2d(periodic):
    """ test grids with singular dimensions """
    dim = np.random.randint(3, 5)
    g1 = UnitGrid([dim], periodic=periodic)
    g2a = UnitGrid([dim, 1], periodic=periodic)
    g2b = UnitGrid([1, dim], periodic=periodic)

    data = np.random.random(dim)
    expected = g1.get_operator("laplace", "natural")(data)
    for g in [g2a, g2b]:
        res = g.get_operator("laplace", "natural")(data.reshape(g.shape))
        np.testing.assert_allclose(expected.flat, res.flat)


@pytest.mark.parametrize("periodic", [True, False])
def test_singular_dimensions_3d(periodic):
    """ test grids with singular dimensions """
    dim = np.random.randint(3, 5)
    g1 = UnitGrid([dim], periodic=periodic)
    g3a = UnitGrid([dim, 1, 1], periodic=periodic)
    g3b = UnitGrid([1, 1, dim], periodic=periodic)

    data = np.random.random(dim)
    expected = g1.get_operator("laplace", "natural")(data)
    for g in [g3a, g3b]:
        res = g.get_operator("laplace", "natural")(data.reshape(g.shape))
        np.testing.assert_allclose(expected.flat, res.flat)


def test_laplace_1d():
    """ test the implementation of the laplace operator """
    for periodic in [True, False]:
        bcs = _get_random_grid_bcs(1, periodic=periodic)
        a = np.random.random(bcs.grid.shape)  # test data
        l1 = ops.make_laplace(bcs, method="scipy")
        l2 = ops.make_laplace(bcs, method="numba")
        l3 = ops.make_laplace(bcs, method="matrix")
        np.testing.assert_allclose(l1(a), l2(a))
        np.testing.assert_allclose(l1(a), l3(a))


def test_laplace_2d():
    """ test the implementation of the laplace operator """
    for periodic in [True, False]:
        bcs = _get_random_grid_bcs(2, dx="uniform", periodic=periodic)
        a = np.random.random(bcs.grid.shape)  # test data

        dx = bcs._uniform_discretization
        kernel = np.array([[0, 1, 0], [1, -4, 1], [0, 1, 0]]) / dx ** 2
        res = ndimage.convolve(a, kernel, **bcs._scipy_border_mode)
        l1 = ops.make_laplace(bcs, method="scipy")
        np.testing.assert_allclose(l1(a), res)

        l2 = ops.make_laplace(bcs, method="numba")
        np.testing.assert_allclose(l2(a), res)

        l3 = ops.make_laplace(bcs, method="matrix")
        np.testing.assert_allclose(l3(a), res)


def test_laplace_2d_nonuniform():
    """test the implementation of the laplace operator for
    non-uniform coordinates"""
    for periodic in [True, False]:
        bcs = _get_random_grid_bcs(ndim=2, dx="random", periodic=periodic)

        dx = bcs.grid.discretization
        kernel_x = np.array([1, -2, 1]) / dx[0] ** 2
        kernel_y = np.array([1, -2, 1]) / dx[1] ** 2
        a = np.random.random(bcs.grid.shape)

        res = ndimage.convolve1d(a, kernel_x, axis=0, **bcs._scipy_border_mode)
        res += ndimage.convolve1d(a, kernel_y, axis=1, **bcs._scipy_border_mode)

        lap = ops._make_laplace_numba_2d(bcs)
        np.testing.assert_allclose(lap(a), res)

        lap = ops.make_laplace(bcs, method="matrix")
        np.testing.assert_allclose(lap(a), res)


def test_laplace_3d():
    """ test the implementation of the laplace operator """
    for periodic in [True, False]:
        bcs = _get_random_grid_bcs(ndim=3, dx="uniform", periodic=periodic)
        a = np.random.random(bcs.grid.shape)
        l1 = ops.make_laplace(bcs, method="scipy")
        l2 = ops.make_laplace(bcs, method="numba")
        np.testing.assert_allclose(l1(a), l2(a))


def test_gradient_1d():
    """ test specific boundary conditions for the 1d gradient """
    grid = UnitGrid(5)

    b_l = {"type": "derivative", "value": -1}
    b_r = {"type": "derivative", "value": 1}
    bcs = grid.get_boundary_conditions([b_l, b_r])
    grad = ops._make_gradient_numba_1d(bcs)
    np.testing.assert_allclose(grad(np.arange(5)), np.ones((1, 5)))

    b_l = {"type": "value", "value": 3}
    b_r = {"type": "value", "value": 3}
    bcs = grid.get_boundary_conditions([b_l, b_r])
    grad = ops._make_gradient_numba_1d(bcs)
    np.testing.assert_allclose(grad(np.full(5, 3)), np.zeros((1, 5)))


@pytest.mark.parametrize("ndim", [1, 2, 3])
def test_gradient(ndim):
    """ test different gradient operators """
    for periodic in [True, False]:
        bcs = _get_random_grid_bcs(ndim, dx="uniform", periodic=periodic)
        op1 = ops.make_gradient(bcs, method="scipy")
        op2 = ops.make_gradient(bcs, method="numba")
        arr = np.random.random(bcs.grid.shape)
        val1, val2 = op1(arr), op2(arr)
        assert val1.shape == (ndim,) + bcs.grid.shape
        np.testing.assert_allclose(val1, val2)


@pytest.mark.parametrize("ndim", [1, 2, 3])
def test_divergence(ndim):
    """ test different divergence operators """
    for periodic in [True, False]:
        bcs = _get_random_grid_bcs(ndim, dx="uniform", periodic=periodic)
        op1 = ops.make_divergence(bcs, method="scipy")
        op2 = ops.make_divergence(bcs, method="numba")
        arr = np.random.random(np.r_[ndim, bcs.grid.shape])
        np.testing.assert_allclose(op1(arr), op2(arr))


@pytest.mark.parametrize("ndim", [1, 2, 3])
def test_vector_gradient(ndim):
    """ test different vector gradient operators """
    bcs = _get_random_grid_bcs(ndim, dx="uniform", periodic="random", rank=1)
    mvg = ops.make_vector_gradient
    op1 = mvg(bcs, method="scipy")
    op2 = mvg(bcs, method="numba")
    arr = np.random.random((ndim,) + bcs.grid.shape)
    val1, val2 = op1(arr), op2(arr)
    assert val1.shape == (ndim, ndim) + bcs.grid.shape
    np.testing.assert_allclose(val1, val2)


@pytest.mark.parametrize("ndim", [1, 2, 3])
def test_vector_laplace(ndim):
    """ test different vector laplace operators """
    bcs = _get_random_grid_bcs(ndim, dx="uniform", periodic="random", rank=1)
    mvg = ops.make_vector_laplace
    op1 = mvg(bcs, method="scipy")
    op2 = mvg(bcs, method="numba")
    arr = np.random.random((ndim,) + bcs.grid.shape)
    val1, val2 = op1(arr), op2(arr)
    assert val1.shape == (ndim,) + bcs.grid.shape
    np.testing.assert_allclose(val1, val2)


@pytest.mark.parametrize("ndim", [1, 2, 3])
def test_tensor_divergence(ndim):
    """ test different tensor divergence operators """
    bcs = _get_random_grid_bcs(ndim, dx="uniform", periodic="random", rank=1)
    op1 = ops.make_tensor_divergence(bcs, method="scipy")
    op2 = ops.make_tensor_divergence(bcs, method="numba")
    arr = np.random.random((ndim, ndim) + bcs.grid.shape)
    val1, val2 = op1(arr), op2(arr)
    assert val1.shape == (ndim,) + bcs.grid.shape
    np.testing.assert_allclose(val1, val2)


def test_div_grad_const():
    """ compare div grad to laplace operator """
    grid = CartesianGrid([[-1, 1]], 32)

    # test constant
    y = ScalarField(grid, 3)
    for bc in [{"type": "derivative", "value": 0}, {"type": "value", "value": 3}]:
        bcs = grid.get_boundary_conditions(bc)
        lap = y.laplace(bcs)
        divgrad = y.gradient(bcs).divergence(bcs.differentiated)
        np.testing.assert_allclose(lap.data, np.zeros(32))
        np.testing.assert_allclose(divgrad.data, np.zeros(32))


def test_div_grad_linear():
    """ compare div grad to laplace operator """
    grid = CartesianGrid([[-1, 1]], 32)
    x = grid.axes_coords[0]

    # test linear
    f = np.random.random() + 1
    y = ScalarField(grid, f * x)

    b1 = [{"type": "neumann", "value": -f}, {"type": "neumann", "value": f}]
    b2 = [{"type": "value", "value": -f}, {"type": "value", "value": f}]
    for bs in [b1, b2]:
        bcs = y.grid.get_boundary_conditions(bs)
        lap = y.laplace(bcs)
        divgrad = y.gradient(bcs).divergence(bcs.differentiated)
        np.testing.assert_allclose(lap.data, np.zeros(32), atol=1e-10)
        np.testing.assert_allclose(divgrad.data, np.zeros(32), atol=1e-10)


def test_div_grad_quadratic():
    """ compare div grad to laplace operator """
    grid = CartesianGrid([[-1, 1]], 32)
    x = grid.axes_coords[0]

    # test simple quadratic
    y = ScalarField(grid, x ** 2)

    bcs = grid.get_boundary_conditions({"type": "derivative", "value": 2})
    lap = y.laplace(bcs)
    divgrad = y.gradient(bcs).divergence(bcs.differentiated)

    np.testing.assert_allclose(lap.data, np.full(32, 2.0))
    np.testing.assert_allclose(divgrad.data, np.full(32, 2.0))


@pytest.mark.parametrize("dim", [1, 2, 3])
def test_gradient_squared(dim):
    """ compare gradient squared operator """
    grid = CartesianGrid(
        [[0, 2 * np.pi]] * dim,
        shape=np.random.randint(30, 35, dim),
        periodic=np.random.choice([False, True], dim),
    )
    field = ScalarField.random_harmonic(grid, modes=1)
    s1 = field.gradient("natural").to_scalar("squared_sum")
    s2 = field.gradient_squared("natural", central=True)
    np.testing.assert_allclose(s1.data, s2.data, rtol=0.1, atol=0.1)
    s3 = field.gradient_squared("natural", central=False)
    np.testing.assert_allclose(s1.data, s3.data, rtol=0.2, atol=0.2)
    assert not np.array_equal(s2.data, s3.data)


def test_rect_div_grad():
    """ compare div grad to laplacian """
    grid = CartesianGrid([[0, 2 * np.pi], [0, 2 * np.pi]], [16, 16], periodic=True)
    x, y = grid.cell_coords[..., 0], grid.cell_coords[..., 1]
    arr = np.cos(x) + np.sin(y)

    bcs = grid.get_boundary_conditions("natural")
    laplace = grid.get_operator("laplace", bcs)
    grad = grid.get_operator("gradient", bcs)
    div = grid.get_operator("divergence", bcs.differentiated)
    a = laplace(arr)
    b = div(grad(arr))
    np.testing.assert_allclose(a, -arr, rtol=0.05, atol=0.01)
    np.testing.assert_allclose(b, -arr, rtol=0.05, atol=0.01)


def test_degenerated_grid():
    """ test operators on grids with singular dimensions """
    g1 = CartesianGrid([[0, 1]], 4)
    g2 = CartesianGrid([[0, 1], [0, 0.1]], [4, 1], periodic=[False, True])
    d = np.random.random(4)

    v1 = g1.get_operator("laplace", bc="natural")(d)
    v2 = g2.get_operator("laplace", bc="natural")(d.reshape(g2.shape))
    assert v2.shape == g2.shape

    np.testing.assert_allclose(v1.flat, v2.flat)


@pytest.mark.parametrize("dim", [1, 2])
def test_poisson_solver_general(dim):
    """ test the poisson solver on Cartesian grids """
    bcs = _get_random_grid_bcs(dim)

    poisson = bcs.grid.get_operator("poisson_solver", bcs)
    laplace = bcs.grid.get_operator("laplace", bcs)

    d = np.random.random(bcs.grid.shape)
    d -= d.mean()  # balance the right hand side
    np.testing.assert_allclose(
        laplace(poisson(d)), d, atol=1e-5, rtol=1e-5, err_msg=f"bcs = {bcs}"
    )


def test_2nd_order_bc():
    """ test whether 2nd order boundary conditions can be used """
    grid = UnitGrid([8, 8])
    field = ScalarField.random_uniform(grid)
    field.laplace([{"value": "sin(y)"}, {"value": "x"}])


@pytest.mark.parametrize("ndim,axis", [(1, 0), (2, 0), (2, 1), (3, 0), (3, 1), (3, 2)])
def test_make_derivative(ndim, axis):
    """ test the make derivative function """
    periodic = random.choice([True, False])
    grid = CartesianGrid([[0, 6 * np.pi]] * ndim, 16, periodic=periodic)
    field = ScalarField.random_harmonic(grid, modes=1, axis_combination=np.add)

    bcs = grid.get_boundary_conditions("natural")
    grad = field.gradient(bcs)
    for method in ["central", "forward", "backward"]:
        msg = f"method={method}, periodic={periodic}"
        diff = ops._make_derivative(bcs, axis=axis, method=method)
        np.testing.assert_allclose(
            grad.data[axis], diff(field.data), atol=0.1, rtol=0.1, err_msg=msg
        )



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np

from pde import CartesianGrid, DiffusionPDE, ScalarField, UnitGrid


def test_diffusion_single():
    """ test some methods of the simple diffusion model """
    eq = DiffusionPDE()
    assert isinstance(str(eq), str)
    assert isinstance(repr(eq), str)
    assert not eq.explicit_time_dependence

    grid = UnitGrid([4, 4])
    state = ScalarField.random_uniform(grid)

    field = eq.evolution_rate(state)
    assert isinstance(field, ScalarField)
    assert field.grid == grid


def test_simple_diffusion_value():
    """ test a simple diffusion equation with constant boundaries """
    grid = CartesianGrid([[0, 1]], [16])
    c = ScalarField.random_uniform(grid, 0, 1)
    b_l = {"type": "value", "value": 0}
    b_r = {"type": "value", "value": 1}
    pde = DiffusionPDE(bc=[b_l, b_r])
    sol, info = pde.solve(c, t_range=1, dt=0.001, tracker=None, ret_info=True)
    assert isinstance(info, dict)
    np.testing.assert_allclose(sol.data, grid.axes_coords[0], rtol=5e-3)


def test_simple_diffusion_flux_right():
    """ test a simple diffusion equation with flux boundary on the right """
    grid = CartesianGrid([[0, 1]], [16])
    c = ScalarField.random_uniform(grid, 0, 1)
    b_l = {"type": "value", "value": 0}
    b_r = {"type": "derivative", "value": 3}
    pde = DiffusionPDE(bc=[b_l, b_r])
    sol = pde.solve(c, t_range=5, dt=0.001, tracker=None)
    np.testing.assert_allclose(sol.data, 3 * grid.axes_coords[0], rtol=5e-3)


def test_simple_diffusion_flux_left():
    """ test a simple diffusion equation with flux boundary on the left """
    grid = CartesianGrid([[0, 1]], [16])
    c = ScalarField.random_uniform(grid, 0, 1)
    b_l = {"type": "derivative", "value": 2}
    b_r = {"type": "value", "value": 0}
    pde = DiffusionPDE(bc=[b_l, b_r])
    sol = pde.solve(c, t_range=5, dt=0.001, tracker=None)
    np.testing.assert_allclose(sol.data, 2 - 2 * grid.axes_coords[0], rtol=5e-3)



"""
This module implements differential operators on Cartesian grids 

.. autosummary::
   :nosignatures:

   make_laplace
   make_gradient
   make_divergence
   make_vector_gradient
   make_vector_laplace
   make_tensor_divergence
   
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>   
"""

from typing import Tuple

import numba as nb
import numpy as np

from ... import config
from ...tools.docstrings import fill_in_docstring
from ...tools.numba import jit_allocate_out
from ...tools.typing import OperatorType
from ..boundaries import Boundaries
from ..cartesian import CartesianGridBase
from .common import make_general_poisson_solver, make_laplace_from_matrix


@fill_in_docstring
def _get_laplace_matrix_1d(bcs: Boundaries) -> Tuple[np.ndarray, np.ndarray]:
    """get sparse matrix for laplace operator on a 1d Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        tuple: A sparse matrix and a sparse vector that can be used to evaluate
        the discretized laplacian
    """
    from scipy import sparse

    dim_x = bcs.grid.shape[0]
    matrix = sparse.dok_matrix((dim_x, dim_x))
    vector = sparse.dok_matrix((dim_x, 1))

    for i in range(dim_x):
        matrix[i, i] += -2

        if i == 0:
            const, entries = bcs[0].get_data((-1,))
            vector[i] += const
            for k, v in entries.items():
                matrix[i, k] += v
        else:
            matrix[i, i - 1] += 1

        if i == dim_x - 1:
            const, entries = bcs[0].get_data((dim_x,))
            vector[i] += const
            for k, v in entries.items():
                matrix[i, k] += v
        else:
            matrix[i, i + 1] += 1

    matrix *= bcs.grid.discretization[0] ** -2
    vector *= bcs.grid.discretization[0] ** -2

    return matrix, vector


@fill_in_docstring
def _get_laplace_matrix_2d(bcs: Boundaries) -> Tuple[np.ndarray, np.ndarray]:
    """get sparse matrix for laplace operator on a 2d Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        tuple: A sparse matrix and a sparse vector that can be used to evaluate
        the discretized laplacian
    """
    from scipy import sparse

    dim_x, dim_y = bcs.grid.shape
    matrix = sparse.dok_matrix((dim_x * dim_y, dim_x * dim_y))
    vector = sparse.dok_matrix((dim_x * dim_y, 1))

    bc_x, bc_y = bcs
    scale_x, scale_y = bcs.grid.discretization ** -2

    def i(x, y):
        """helper function for flattening the index

        This is equivalent to np.ravel_multi_index((x, y), (dim_x, dim_y))
        """
        return x * dim_y + y

    # set diagonal elements, i.e., the central value in the kernel
    matrix.setdiag(-2 * (scale_x + scale_y))

    for x in range(dim_x):
        for y in range(dim_y):
            # handle x-direction
            if x == 0:
                const, entries = bc_x.get_data((-1, y))
                vector[i(x, y)] += const * scale_x
                for k, v in entries.items():
                    matrix[i(x, y), i(k, y)] += v * scale_x
            else:
                matrix[i(x, y), i(x - 1, y)] += scale_x

            if x == dim_x - 1:
                const, entries = bc_x.get_data((dim_x, y))
                vector[i(x, y)] += const * scale_x
                for k, v in entries.items():
                    matrix[i(x, y), i(k, y)] += v * scale_x
            else:
                matrix[i(x, y), i(x + 1, y)] += scale_x

            # handle y-direction
            if y == 0:
                const, entries = bc_y.get_data((x, -1))
                vector[i(x, y)] += const * scale_y
                for k, v in entries.items():
                    matrix[i(x, y), i(x, k)] += v * scale_y
            else:
                matrix[i(x, y), i(x, y - 1)] += scale_y

            if y == dim_y - 1:
                const, entries = bc_y.get_data((x, dim_y))
                vector[i(x, y)] += const * scale_y
                for k, v in entries.items():
                    matrix[i(x, y), i(x, k)] += v * scale_y
            else:
                matrix[i(x, y), i(x, y + 1)] += scale_y

    return matrix, vector


@fill_in_docstring
def _get_laplace_matrix(bcs: Boundaries) -> Tuple[np.ndarray, np.ndarray]:
    """get sparse matrix for laplace operator on a 1d Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        tuple: A sparse matrix and a sparse vector that can be used to evaluate
        the discretized laplacian
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(0)

    if dim == 1:
        result = _get_laplace_matrix_1d(bcs)
    elif dim == 2:
        result = _get_laplace_matrix_2d(bcs)
    else:
        raise NotImplementedError(
            f"Numba laplace operator not implemented for {dim:d} dimensions"
        )

    return result


def _make_derivative(
    bcs: Boundaries, axis: int = 0, method: str = "central"
) -> OperatorType:
    """make a derivative operator along a single axis using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        axis (int):
            The axis along which the derivative will be taken
        method (str):
            The method for calculating the derivative. Possible values are
            'central', 'forward', and 'backward'.

    Returns:
        A function that can be applied to an array of values. The result will be
        an array of the same shape containing the actual derivatives at the grid
        points.
    """
    if method not in {"central", "forward", "backward"}:
        raise ValueError(f"Unknown derivative type `{method}`")

    shape = bcs.grid.shape
    dim = len(shape)
    dx = bcs.grid.discretization[axis]
    region = bcs[axis].make_region_evaluator()

    if dim == 1:

        @jit_allocate_out(out_shape=shape)
        def diff(arr, out=None):
            """ calculate derivative of 1d array `arr` """
            for i in range(shape[0]):
                arr_l, arr_m, arr_h = region(arr, (i,))
                if method == "central":
                    out[i] = (arr_h - arr_l) * 0.5 / dx
                elif method == "forward":
                    out[i] = (arr_h - arr_m) / dx
                elif method == "backward":
                    out[i] = (arr_m - arr_l) / dx

            return out

    elif dim == 2:

        @jit_allocate_out(out_shape=shape)
        def diff(arr, out=None):
            """ calculate derivative of 2d array `arr` """
            for i in range(shape[0]):
                for j in range(shape[1]):
                    arr_l, arr_m, arr_h = region(arr, (i, j))
                    if method == "central":
                        out[i, j] = (arr_h - arr_l) * 0.5 / dx
                    elif method == "forward":
                        out[i, j] = (arr_h - arr_m) / dx
                    elif method == "backward":
                        out[i, j] = (arr_m - arr_l) / dx

            return out

    elif dim == 3:

        @jit_allocate_out(out_shape=shape)
        def diff(arr, out=None):
            """ calculate derivative of 3d array `arr` """
            for i in range(shape[0]):
                for j in range(shape[1]):
                    for k in range(shape[2]):
                        arr_l, arr_m, arr_h = region(arr, (i, j, k))
                        if method == "central":
                            out[i, j, k] = (arr_h - arr_l) * 0.5 / dx
                        elif method == "forward":
                            out[i, j, k] = (arr_h - arr_m) / dx
                        elif method == "backward":
                            out[i, j, k] = (arr_m - arr_l) / dx

            return out

    else:
        raise NotImplementedError(
            f"Numba derivative operator not implemented for {dim:d} dimensions"
        )

    return diff  # type: ignore


@fill_in_docstring
def _make_laplace_scipy_nd(bcs: Boundaries) -> OperatorType:
    """make a laplace operator using the scipy module

    This only supports uniform discretizations.

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    from scipy import ndimage

    scaling = bcs._uniform_discretization ** -2
    args = bcs._scipy_border_mode

    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        assert arr.shape == bcs.grid.shape
        return ndimage.laplace(scaling * arr, output=out, **args)

    return laplace


@fill_in_docstring
def _make_laplace_numba_1d(bcs: Boundaries) -> OperatorType:
    """make a 1d laplace operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    dim_x = bcs.grid.shape[0]
    scale = bcs.grid.discretization[0] ** -2
    region_x = bcs[0].make_region_evaluator()

    @jit_allocate_out
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        for i in range(dim_x):
            valm, val, valp = region_x(arr, (i,))
            out[i] = (valm - 2 * val + valp) * scale

        return out

    return laplace  # type: ignore


@fill_in_docstring
def _make_laplace_numba_2d(bcs: Boundaries) -> OperatorType:
    """make a 2d laplace operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    dim_x, dim_y = bcs.grid.shape
    scale_x, scale_y = bcs.grid.discretization ** -2

    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()

    # use parallel processing for large enough arrays
    parallel = dim_x * dim_y >= config["numba.parallel_threshold"]

    @jit_allocate_out(parallel=parallel)
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        for i in nb.prange(dim_x):

            # left-most part
            j = 0
            arr_x_l, arr_c, arr_x_h = region_x(arr, (i, j))
            arr_y_l, _, arr_y_h = region_y(arr, (i, j))
            lap_x = (arr_x_l - 2 * arr_c + arr_x_h) * scale_x
            lap_y = (arr_y_l - 2 * arr_c + arr_y_h) * scale_y
            out[i, j] = lap_x + lap_y

            if dim_y == 1:
                continue  # deal with singular y dimension

            # middle part without boundary conditions in y-direction
            for j in range(1, dim_y - 1):
                arr_x_l, arr_c, arr_x_h = region_x(arr, (i, j))
                arr_y_l = arr[i, j - 1]
                arr_y_h = arr[i, j + 1]

                lap_x = (arr_x_l - 2 * arr_c + arr_x_h) * scale_x
                lap_y = (arr_y_l - 2 * arr_c + arr_y_h) * scale_y
                out[i, j] = lap_x + lap_y

            # right-most part
            j = dim_y - 1
            arr_x_l, arr_c, arr_x_h = region_x(arr, (i, j))
            arr_y_l, _, arr_y_h = region_y(arr, (i, j))
            lap_x = (arr_x_l - 2 * arr_c + arr_x_h) * scale_x
            lap_y = (arr_y_l - 2 * arr_c + arr_y_h) * scale_y
            out[i, j] = lap_x + lap_y

        return out

    return laplace  # type: ignore


@fill_in_docstring
def _make_laplace_numba_3d(bcs: Boundaries) -> OperatorType:
    """make a 3d laplace operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    dim_x, dim_y, dim_z = bcs.grid.shape
    scale_x, scale_y, scale_z = bcs.grid.discretization ** -2

    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()
    region_z = bcs[2].make_region_evaluator()

    # use parallel processing for large enough arrays
    parallel = dim_x * dim_y * dim_z >= config["numba.parallel_threshold"]

    @jit_allocate_out(parallel=parallel)
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        for i in nb.prange(dim_x):
            for j in range(dim_y):
                for k in range(dim_z):

                    arr_x_l, arr_c, arr_x_h = region_x(arr, (i, j, k))
                    arr_y_l, _, arr_y_h = region_y(arr, (i, j, k))
                    arr_z_l, _, arr_z_h = region_z(arr, (i, j, k))

                    lap_x = (arr_x_l - 2 * arr_c + arr_x_h) * scale_x
                    lap_y = (arr_y_l - 2 * arr_c + arr_y_h) * scale_y
                    lap_z = (arr_z_l - 2 * arr_c + arr_z_h) * scale_z
                    out[i, j, k] = lap_x + lap_y + lap_z

        return out

    return laplace  # type: ignore


@CartesianGridBase.register_operator("laplace", rank_in=0, rank_out=0)
@fill_in_docstring
def make_laplace(bcs: Boundaries, method: str = "auto") -> OperatorType:
    """make a laplace operator on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the laplace operator.
            If method='auto', a suitable method is chosen automatically.

    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(0)

    if method == "auto":
        # choose the fastest available Laplace operator
        if 1 <= dim <= 3:
            method = "numba"
        else:
            method = "scipy"

    if method == "numba":
        if dim == 1:
            laplace = _make_laplace_numba_1d(bcs)
        elif dim == 2:
            laplace = _make_laplace_numba_2d(bcs)
        elif dim == 3:
            laplace = _make_laplace_numba_3d(bcs)
        else:
            raise NotImplementedError(
                f"Numba laplace operator not implemented for {dim:d} dimensions"
            )

    elif method == "matrix":
        laplace = make_laplace_from_matrix(*_get_laplace_matrix(bcs))

    elif method == "scipy":
        laplace = _make_laplace_scipy_nd(bcs)

    else:
        raise ValueError(f"Method `{method}` is not defined")

    return laplace


@fill_in_docstring
def _make_gradient_scipy_nd(bcs: Boundaries) -> OperatorType:
    """make a gradient operator using the scipy module

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    from scipy import ndimage

    scaling = 0.5 / bcs.grid.discretization
    dim = bcs.grid.dim
    shape_out = (dim,) + bcs.grid.shape
    args = bcs._scipy_border_mode

    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        assert arr.shape == bcs.grid.shape
        if out is None:
            out = np.empty(shape_out)
        else:
            assert out.shape == shape_out

        for i in range(dim):
            out[i] = ndimage.convolve1d(arr, [1, 0, -1], axis=i, **args) * scaling[i]
        return out

    return gradient


@fill_in_docstring
def _make_gradient_numba_1d(bcs: Boundaries) -> OperatorType:
    """make a 1d gradient operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    dim_x = bcs.grid.shape[0]
    scale = 0.5 / bcs.grid.discretization[0]
    region_x = bcs[0].make_region_evaluator()

    @jit_allocate_out(out_shape=(1, dim_x))
    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in range(dim_x):
            valm, _, valp = region_x(arr, (i,))
            out[0, i] = (valp - valm) * scale

        return out

    return gradient  # type: ignore


@fill_in_docstring
def _make_gradient_numba_2d(bcs: Boundaries) -> OperatorType:
    """make a 2d gradient operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    dim_x, dim_y = bcs.grid.shape
    scale_x, scale_y = 0.5 / bcs.grid.discretization

    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()

    # use parallel processing for large enough arrays
    parallel = dim_x * dim_y >= config["numba.parallel_threshold"]

    @jit_allocate_out(parallel=parallel, out_shape=(2, dim_x, dim_y))
    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in nb.prange(dim_x):
            for j in range(dim_y):
                arr_x_l, _, arr_x_h = region_x(arr, (i, j))
                arr_y_l, _, arr_y_h = region_y(arr, (i, j))

                out[0, i, j] = (arr_x_h - arr_x_l) * scale_x
                out[1, i, j] = (arr_y_h - arr_y_l) * scale_y

        return out

    return gradient  # type: ignore


@fill_in_docstring
def _make_gradient_numba_3d(bcs: Boundaries) -> OperatorType:
    """make a 3d gradient operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    dim_x, dim_y, dim_z = bcs.grid.shape
    scale_x, scale_y, scale_z = 0.5 / bcs.grid.discretization

    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()
    region_z = bcs[2].make_region_evaluator()

    # use parallel processing for large enough arrays
    parallel = dim_x * dim_y * dim_z >= config["numba.parallel_threshold"]

    @jit_allocate_out(parallel=parallel, out_shape=(3, dim_x, dim_y, dim_z))
    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in nb.prange(dim_x):
            for j in range(dim_y):
                for k in range(dim_z):
                    arr_x_l, _, arr_x_h = region_x(arr, (i, j, k))
                    arr_y_l, _, arr_y_h = region_y(arr, (i, j, k))
                    arr_z_l, _, arr_z_h = region_z(arr, (i, j, k))

                    out[0, i, j, k] = (arr_x_h - arr_x_l) * scale_x
                    out[1, i, j, k] = (arr_y_h - arr_y_l) * scale_y
                    out[2, i, j, k] = (arr_z_h - arr_z_l) * scale_z

        return out

    return gradient  # type: ignore


@CartesianGridBase.register_operator("gradient", rank_in=0, rank_out=1)
@fill_in_docstring
def make_gradient(bcs: Boundaries, method: str = "auto") -> OperatorType:
    """make a gradient operator on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the gradient operator.
            If method='auto', a suitable method is chosen automatically.

    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(0)

    if method == "auto":
        # choose the fastest available gradient operator
        if 1 <= dim <= 3:
            method = "numba"
        else:
            method = "scipy"

    if method == "numba":
        if dim == 1:
            gradient = _make_gradient_numba_1d(bcs)
        elif dim == 2:
            gradient = _make_gradient_numba_2d(bcs)
        elif dim == 3:
            gradient = _make_gradient_numba_3d(bcs)
        else:
            raise NotImplementedError(
                f"Numba gradient operator not implemented for dimension {dim}"
            )

    elif method == "scipy":
        gradient = _make_gradient_scipy_nd(bcs)

    else:
        raise ValueError(f"Method `{method}` is not defined")

    return gradient


@fill_in_docstring
def _make_gradient_squared_numba_1d(
    bcs: Boundaries, central: bool = True
) -> OperatorType:
    """make a 1d squared gradient operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        central (bool):
            Whether a central difference approximation is used for the gradient
            operator. If this is False, the squared gradient is calculated as
            the mean of the squared values of the forward and backward
            derivatives.

    Returns:
        A function that can be applied to an array of values
    """
    dim_x = bcs.grid.shape[0]
    region_x = bcs[0].make_region_evaluator()

    if central:
        # use central differences
        scale = 1 / (2 * bcs.grid.discretization[0]) ** 2

        @jit_allocate_out(out_shape=(dim_x,))
        def gradient_squared(arr, out=None):
            """ apply squared gradient operator to array `arr` """
            for i in range(dim_x):
                valm, _, valp = region_x(arr, (i,))
                out[i] = (valp - valm) ** 2 * scale

            return out

    else:
        # use forward and backward differences
        scale = 1 / (2 * bcs.grid.discretization[0] ** 2)

        @jit_allocate_out(out_shape=(dim_x,))
        def gradient_squared(arr, out=None):
            """ apply squared gradient operator to array `arr` """
            for i in range(dim_x):
                valm, valc, valp = region_x(arr, (i,))
                out[i] = ((valp - valc) ** 2 + (valc - valm) ** 2) * scale

            return out

    return gradient_squared  # type: ignore


@fill_in_docstring
def _make_gradient_squared_numba_2d(
    bcs: Boundaries, central: bool = True
) -> OperatorType:
    """make a 2d squared gradient operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        central (bool):
            Whether a central difference approximation is used for the gradient
            operator. If this is False, the squared gradient is calculated as
            the mean of the squared values of the forward and backward
            derivatives.

    Returns:
        A function that can be applied to an array of values
    """
    dim_x, dim_y = bcs.grid.shape

    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()

    # use parallel processing for large enough arrays
    parallel = dim_x * dim_y >= config["numba.parallel_threshold"]

    if central:
        # use central differences
        scale_x, scale_y = 1 / (2 * bcs.grid.discretization) ** 2

        @jit_allocate_out(parallel=parallel, out_shape=(dim_x, dim_y))
        def gradient_squared(arr, out=None):
            """ apply squared gradient operator to array `arr` """
            for i in nb.prange(dim_x):
                for j in range(dim_y):
                    arr_x_l, _, arr_x_h = region_x(arr, (i, j))
                    arr_y_l, _, arr_y_h = region_y(arr, (i, j))

                    term_x = (arr_x_h - arr_x_l) ** 2 * scale_x
                    term_y = (arr_y_h - arr_y_l) ** 2 * scale_y
                    out[i, j] = term_x + term_y

            return out

    else:
        # use forward and backward differences
        scale_x, scale_y = 1 / (2 * bcs.grid.discretization ** 2)

        @jit_allocate_out(parallel=parallel, out_shape=(dim_x, dim_y))
        def gradient_squared(arr, out=None):
            """ apply squared gradient operator to array `arr` """
            for i in nb.prange(dim_x):
                for j in range(dim_y):
                    arr_x_l, arr_c, arr_x_h = region_x(arr, (i, j))
                    arr_y_l, _, arr_y_h = region_y(arr, (i, j))

                    term_x = ((arr_x_h - arr_c) ** 2 + (arr_c - arr_x_l) ** 2) * scale_x
                    term_y = ((arr_y_h - arr_c) ** 2 + (arr_c - arr_y_l) ** 2) * scale_y
                    out[i, j] = term_x + term_y

            return out

    return gradient_squared  # type: ignore


@fill_in_docstring
def _make_gradient_squared_numba_3d(
    bcs: Boundaries, central: bool = True
) -> OperatorType:
    """make a 3d squared gradient operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        central (bool):
            Whether a central difference approximation is used for the gradient
            operator. If this is False, the squared gradient is calculated as
            the mean of the squared values of the forward and backward
            derivatives.

    Returns:
        A function that can be applied to an array of values
    """
    dim_x, dim_y, dim_z = bcs.grid.shape

    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()
    region_z = bcs[2].make_region_evaluator()

    # use parallel processing for large enough arrays
    parallel = dim_x * dim_y * dim_z >= config["numba.parallel_threshold"]

    if central:
        # use central differences
        scale_x, scale_y, scale_z = 1 / (2 * bcs.grid.discretization) ** 2

        @jit_allocate_out(parallel=parallel, out_shape=(dim_x, dim_y, dim_z))
        def gradient_squared(arr, out=None):
            """ apply squared gradient operator to array `arr` """
            for i in nb.prange(dim_x):
                for j in range(dim_y):
                    for k in range(dim_z):
                        arr_x_l, _, arr_x_h = region_x(arr, (i, j, k))
                        arr_y_l, _, arr_y_h = region_y(arr, (i, j, k))
                        arr_z_l, _, arr_z_h = region_z(arr, (i, j, k))

                        term_x = (arr_x_h - arr_x_l) ** 2 * scale_x
                        term_y = (arr_y_h - arr_y_l) ** 2 * scale_y
                        term_z = (arr_z_h - arr_z_l) ** 2 * scale_z
                        out[i, j, k] = term_x + term_y + term_z

            return out

    else:
        # use forward and backward differences
        scale_x, scale_y, scale_z = 1 / (2 * bcs.grid.discretization ** 2)

        @jit_allocate_out(parallel=parallel, out_shape=(dim_x, dim_y, dim_z))
        def gradient_squared(arr, out=None):
            """ apply squared gradient operator to array `arr` """
            for i in nb.prange(dim_x):
                for j in range(dim_y):
                    for k in range(dim_z):
                        arr_x_l, arr_c, arr_x_h = region_x(arr, (i, j, k))
                        arr_y_l, _, arr_y_h = region_y(arr, (i, j, k))
                        arr_z_l, _, arr_z_h = region_z(arr, (i, j, k))

                        term_x = (arr_x_h - arr_c) ** 2 + (arr_c - arr_x_l) ** 2
                        term_x *= scale_x
                        term_y = (arr_y_h - arr_c) ** 2 + (arr_c - arr_y_l) ** 2
                        term_y *= scale_y
                        term_z = (arr_z_h - arr_c) ** 2 + (arr_z_l - arr_c) ** 2
                        term_z *= scale_z
                        out[i, j, k] = term_x + term_y + term_z

            return out

    return gradient_squared  # type: ignore


@CartesianGridBase.register_operator("gradient_squared", rank_in=0, rank_out=0)
@fill_in_docstring
def make_gradient_squared(bcs: Boundaries, central: bool = True) -> OperatorType:
    """make a gradient operator on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        central (bool):
            Whether a central difference approximation is used for the gradient
            operator. If this is False, the squared gradient is calculated as
            the mean of the squared values of the forward and backward
            derivatives.

    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(0)

    if dim == 1:
        gradient_squared = _make_gradient_squared_numba_1d(bcs, central=central)
    elif dim == 2:
        gradient_squared = _make_gradient_squared_numba_2d(bcs, central=central)
    elif dim == 3:
        gradient_squared = _make_gradient_squared_numba_3d(bcs, central=central)
    else:
        raise NotImplementedError(
            f"Squared gradient operator is not implemented for dimension {dim}"
        )

    return gradient_squared


@fill_in_docstring
def _make_divergence_scipy_nd(bcs: Boundaries) -> OperatorType:
    """make a divergence operator using the scipy module

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    from scipy import ndimage

    shape = bcs.grid.shape
    scaling = 0.5 / bcs.grid.discretization
    args = bcs._scipy_border_mode

    def divergence(arr, out=None):
        """ apply divergence operator to array `arr` """
        assert arr.shape[0] == len(shape) and arr.shape[1:] == shape

        # need to initialize with zeros since data is added later
        if out is None:
            out = np.zeros(arr.shape[1:])
        else:
            out[:] = 0

        for i in range(len(shape)):
            out += ndimage.convolve1d(arr[i], [1, 0, -1], axis=i, **args) * scaling[i]
        return out

    return divergence


@fill_in_docstring
def _make_divergence_numba_1d(bcs: Boundaries) -> OperatorType:
    """make a 1d divergence operator using numba compilation

    Args:
        dim (int): The number of support points for each axes
        boundaries (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        dx (float): The discretization

    Returns:
        A function that can be applied to an array of values
    """
    dim_x = bcs.grid.shape[0]
    scale = 0.5 / bcs.grid.discretization[0]
    region_x = bcs[0].make_region_evaluator()

    @jit_allocate_out(out_shape=(dim_x,))
    def divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in range(dim_x):
            valm, _, valp = region_x(arr[0], (i,))
            out[i] = (valp - valm) * scale

        return out

    return divergence  # type: ignore


@fill_in_docstring
def _make_divergence_numba_2d(bcs: Boundaries) -> OperatorType:
    """make a 2d divergence operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    dim_x, dim_y = bcs.grid.shape
    scale_x, scale_y = 0.5 / bcs.grid.discretization

    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()

    # use parallel processing for large enough arrays
    parallel = dim_x * dim_y >= config["numba.parallel_threshold"]

    @jit_allocate_out(parallel=parallel, out_shape=(dim_x, dim_y))
    def divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in nb.prange(dim_x):
            for j in range(dim_y):
                arr_x_l, _, arr_x_h = region_x(arr[0], (i, j))
                arr_y_l, _, arr_y_h = region_y(arr[1], (i, j))

                d_x = (arr_x_h - arr_x_l) * scale_x
                d_y = (arr_y_h - arr_y_l) * scale_y
                out[i, j] = d_x + d_y

        return out

    return divergence  # type: ignore


@fill_in_docstring
def _make_divergence_numba_3d(bcs: Boundaries) -> OperatorType:
    """make a 3d divergence operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    dim_x, dim_y, dim_z = bcs.grid.shape
    scale_x, scale_y, scale_z = 0.5 / bcs.grid.discretization

    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()
    region_z = bcs[2].make_region_evaluator()

    # use parallel processing for large enough arrays
    parallel = dim_x * dim_y * dim_z >= config["numba.parallel_threshold"]

    @jit_allocate_out(parallel=parallel, out_shape=(dim_x, dim_y, dim_z))
    def divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in nb.prange(dim_x):
            for j in range(dim_y):
                for k in range(dim_z):
                    arr_x_l, _, arr_x_h = region_x(arr[0], (i, j, k))
                    arr_y_l, _, arr_y_h = region_y(arr[1], (i, j, k))
                    arr_z_l, _, arr_z_h = region_z(arr[2], (i, j, k))

                    d_x = (arr_x_h - arr_x_l) * scale_x
                    d_y = (arr_y_h - arr_y_l) * scale_y
                    d_z = (arr_z_h - arr_z_l) * scale_z
                    out[i, j, k] = d_x + d_y + d_z

        return out

    return divergence  # type: ignore


@CartesianGridBase.register_operator("divergence", rank_in=1, rank_out=0)
@fill_in_docstring
def make_divergence(bcs: Boundaries, method: str = "auto") -> OperatorType:
    """make a divergence operator on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the divergence operator.
            If method='auto', a suitable method is chosen automatically.

    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(0)

    if method == "auto":
        # choose the fastest available divergence operator
        if 1 <= dim <= 3:
            method = "numba"
        else:
            method = "scipy"

    if method == "numba":
        if dim == 1:
            divergence = _make_divergence_numba_1d(bcs)
        elif dim == 2:
            divergence = _make_divergence_numba_2d(bcs)
        elif dim == 3:
            divergence = _make_divergence_numba_3d(bcs)
        else:
            raise NotImplementedError(
                f"Numba divergence operator not implemented for dimension {dim}"
            )

    elif method == "scipy":
        divergence = _make_divergence_scipy_nd(bcs)

    else:
        raise ValueError(f"Method `{method}` is not defined")

    return divergence


@fill_in_docstring
def _make_vector_gradient_scipy_nd(bcs: Boundaries) -> OperatorType:
    """make a vector gradient operator using the scipy module

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    from scipy import ndimage

    scaling = 0.5 / bcs.grid.discretization
    args = bcs._scipy_border_mode
    dim = bcs.grid.dim
    shape_out = (dim, dim) + bcs.grid.shape

    def vector_gradient(arr, out=None):
        """ apply vector gradient operator to array `arr` """
        assert arr.shape == shape_out[1:]
        if out is None:
            out = np.empty(shape_out)
        else:
            assert out.shape == shape_out

        for i in range(dim):
            for j in range(dim):
                conv = ndimage.convolve1d(arr[j], [1, 0, -1], axis=i, **args)
                out[i, j] = conv * scaling[i]
        return out

    return vector_gradient


@fill_in_docstring
def _make_vector_gradient_numba_1d(bcs: Boundaries) -> OperatorType:
    """make a 1d vector gradient operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    gradient = _make_gradient_numba_1d(bcs.extract_component(0))

    @jit_allocate_out(out_shape=(1, 1) + bcs.grid.shape)
    def vector_gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        gradient(arr[0], out=out[0])
        return out

    return vector_gradient  # type: ignore


@fill_in_docstring
def _make_vector_gradient_numba_2d(bcs: Boundaries) -> OperatorType:
    """make a 2d vector gradient operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    gradient_x = _make_gradient_numba_2d(bcs.extract_component(0))
    gradient_y = _make_gradient_numba_2d(bcs.extract_component(1))

    @jit_allocate_out(out_shape=(2, 2) + bcs.grid.shape)
    def vector_gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        gradient_x(arr[0], out=out[:, 0])
        gradient_y(arr[1], out=out[:, 1])
        return out

    return vector_gradient  # type: ignore


@fill_in_docstring
def _make_vector_gradient_numba_3d(bcs: Boundaries) -> OperatorType:
    """make a 3d vector gradient operator using numba compilation

    Args:
        shape (tuple): The number of support points for each axes
        boundaries (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        dx (float or list): The discretizations

    Returns:
        A function that can be applied to an array of values
    """
    gradient_x = _make_gradient_numba_3d(bcs.extract_component(0))
    gradient_y = _make_gradient_numba_3d(bcs.extract_component(1))
    gradient_z = _make_gradient_numba_3d(bcs.extract_component(2))

    @jit_allocate_out(out_shape=(3, 3) + bcs.grid.shape)
    def vector_gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        gradient_x(arr[0], out=out[:, 0])
        gradient_y(arr[1], out=out[:, 1])
        gradient_z(arr[2], out=out[:, 2])
        return out

    return vector_gradient  # type: ignore


@CartesianGridBase.register_operator("vector_gradient", rank_in=1, rank_out=2)
@fill_in_docstring
def make_vector_gradient(bcs: Boundaries, method: str = "auto") -> OperatorType:
    """make a vector gradient operator on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the vector gradient operator.
            If method='auto', a suitable method is chosen automatically

    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(1)

    # choose the fastest available vector gradient operator
    if method == "auto":
        if 1 <= dim <= 3:
            method = "numba"
        else:
            method = "scipy"

    if method == "numba":
        if dim == 1:
            gradient = _make_vector_gradient_numba_1d(bcs)
        elif dim == 2:
            gradient = _make_vector_gradient_numba_2d(bcs)
        elif dim == 3:
            gradient = _make_vector_gradient_numba_3d(bcs)
        else:
            raise NotImplementedError(
                f"Numba vector gradient operator not implemented for dimension {dim}"
            )

    elif method == "scipy":
        gradient = _make_vector_gradient_scipy_nd(bcs)
    else:
        raise ValueError(f"Method `{method}` is not defined")

    return gradient


@fill_in_docstring
def _make_vector_laplace_scipy_nd(bcs: Boundaries) -> OperatorType:
    """make a vector Laplacian using the scipy module

    This only supports uniform discretizations.

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    from scipy import ndimage

    scaling = bcs._uniform_discretization ** -2
    args = bcs._scipy_border_mode
    dim = bcs.grid.dim
    shape_out = (dim,) + bcs.grid.shape

    def vector_laplace(arr, out=None):
        """ apply vector Laplacian operator to array `arr` """
        assert arr.shape == shape_out
        if out is None:
            out = np.empty(shape_out)
        else:
            assert out.shape == shape_out

        for i in range(dim):
            ndimage.laplace(arr[i], output=out[i], **args)
        return out * scaling

    return vector_laplace


@fill_in_docstring
def _make_vector_laplace_numba_1d(bcs: Boundaries) -> OperatorType:
    """make a 1d vector Laplacian using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    laplace = _make_laplace_numba_1d(bcs.extract_component(0))

    @jit_allocate_out(out_shape=(1,) + bcs.grid.shape)
    def vector_laplace(arr, out=None):
        """ apply vector Laplacian to array `arr` """
        laplace(arr[0], out=out[0])
        return out

    return vector_laplace  # type: ignore


@fill_in_docstring
def _make_vector_laplace_numba_2d(bcs: Boundaries) -> OperatorType:
    """make a 2d vector Laplacian using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    laplace_x = _make_laplace_numba_2d(bcs.extract_component(0))
    laplace_y = _make_laplace_numba_2d(bcs.extract_component(1))

    @jit_allocate_out(out_shape=(2,) + bcs.grid.shape)
    def vector_laplace(arr, out=None):
        """ apply vector Laplacian  to array `arr` """
        laplace_x(arr[0], out=out[0])
        laplace_y(arr[1], out=out[1])
        return out

    return vector_laplace  # type: ignore


@fill_in_docstring
def _make_vector_laplace_numba_3d(bcs: Boundaries) -> OperatorType:
    """make a 3d vector Laplacian using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    laplace_x = _make_laplace_numba_3d(bcs.extract_component(0))
    laplace_y = _make_laplace_numba_3d(bcs.extract_component(1))
    laplace_z = _make_laplace_numba_3d(bcs.extract_component(2))

    @jit_allocate_out(out_shape=(3,) + bcs.grid.shape)
    def vector_laplace(arr, out=None):
        """ apply vector Laplacian to array `arr` """
        laplace_x(arr[0], out=out[0])
        laplace_y(arr[1], out=out[1])
        laplace_z(arr[2], out=out[2])
        return out

    return vector_laplace  # type: ignore


@CartesianGridBase.register_operator("vector_laplace", rank_in=1, rank_out=1)
@fill_in_docstring
def make_vector_laplace(bcs: Boundaries, method: str = "auto") -> OperatorType:
    """make a vector Laplacian on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the vector laplace operator.
            If method='auto', a suitable method is chosen automatically.

    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(1)

    # choose the fastest available vector gradient operator
    if method == "auto":
        if 1 <= dim <= 3:
            method = "numba"
        else:
            method = "scipy"

    if method == "numba":
        if dim == 1:
            gradient = _make_vector_laplace_numba_1d(bcs)
        elif dim == 2:
            gradient = _make_vector_laplace_numba_2d(bcs)
        elif dim == 3:
            gradient = _make_vector_laplace_numba_3d(bcs)
        else:
            raise NotImplementedError(
                f"Numba vector gradient operator not implemented for dimension {dim}"
            )

    elif method == "scipy":
        gradient = _make_vector_laplace_scipy_nd(bcs)
    else:
        raise ValueError(f"Method `{method}` is not defined")

    return gradient


@fill_in_docstring
def _make_tensor_divergence_scipy_nd(bcs: Boundaries) -> OperatorType:
    """make a tensor divergence operator using the scipy module

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    from scipy import ndimage

    scaling = 0.5 / bcs.grid.discretization
    args = bcs._scipy_border_mode
    dim = bcs.grid.dim
    shape_out = (dim,) + bcs.grid.shape

    def tensor_divergence(arr, out=None):
        """ apply tensor divergence operator to array `arr` """
        # need to initialize with zeros since data is added later
        assert arr.shape[0] == dim and arr.shape[1:] == shape_out
        if out is None:
            out = np.zeros(shape_out)
        else:
            assert out.shape == shape_out
            out[:] = 0

        for i in range(dim):
            for j in range(dim):
                conv = ndimage.convolve1d(arr[i, j], [1, 0, -1], axis=j, **args)
                out[i] += conv * scaling[j]
        return out

    return tensor_divergence


@fill_in_docstring
def _make_tensor_divergence_numba_1d(bcs: Boundaries) -> OperatorType:
    """make a 1d tensor divergence  operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    divergence = _make_divergence_numba_1d(bcs.extract_component(0))

    @jit_allocate_out(out_shape=(1,) + bcs.grid.shape)
    def tensor_divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        divergence(arr[0], out=out[0])
        return out

    return tensor_divergence  # type: ignore


@fill_in_docstring
def _make_tensor_divergence_numba_2d(bcs: Boundaries) -> OperatorType:
    """make a 2d tensor divergence  operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    divergence_x = _make_divergence_numba_2d(bcs.extract_component(0))
    divergence_y = _make_divergence_numba_2d(bcs.extract_component(1))

    @jit_allocate_out(out_shape=(2,) + bcs.grid.shape)
    def tensor_divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        divergence_x(arr[0], out=out[0])
        divergence_y(arr[1], out=out[1])
        return out

    return tensor_divergence  # type: ignore


@fill_in_docstring
def _make_tensor_divergence_numba_3d(bcs: Boundaries) -> OperatorType:
    """make a 3d tensor divergence  operator using numba compilation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}

    Returns:
        A function that can be applied to an array of values
    """
    divergence_x = _make_divergence_numba_3d(bcs.extract_component(0))
    divergence_y = _make_divergence_numba_3d(bcs.extract_component(1))
    divergence_z = _make_divergence_numba_3d(bcs.extract_component(2))

    @jit_allocate_out(out_shape=(3,) + bcs.grid.shape)
    def tensor_divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        divergence_x(arr[0], out=out[0])
        divergence_y(arr[1], out=out[1])
        divergence_z(arr[2], out=out[2])
        return out

    return tensor_divergence  # type: ignore


@CartesianGridBase.register_operator("tensor_divergence", rank_in=2, rank_out=1)
@fill_in_docstring
def make_tensor_divergence(bcs: Boundaries, method: str = "auto") -> OperatorType:
    """make a tensor divergence operator on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the tensor divergence
            operator. If method='auto', a suitable method is chosen
            automatically.

    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(1)

    # choose the fastest available tensor divergence operator
    if method == "auto":
        if 1 <= dim <= 3:
            method = "numba"
        else:
            method = "scipy"

    if method == "numba":
        if dim == 1:
            func = _make_tensor_divergence_numba_1d(bcs)
        elif dim == 2:
            func = _make_tensor_divergence_numba_2d(bcs)
        elif dim == 3:
            func = _make_tensor_divergence_numba_3d(bcs)
        else:
            raise NotImplementedError(
                f"Numba tensor divergence operator not implemented for dimension {dim}"
            )

    elif method == "scipy":
        func = _make_tensor_divergence_scipy_nd(bcs)
    else:
        raise ValueError(f"Method `{method}` is not defined")

    return func


@CartesianGridBase.register_operator("poisson_solver", rank_in=0, rank_out=0)
@fill_in_docstring
def make_poisson_solver(bcs: Boundaries, method: str = "auto") -> OperatorType:
    """make a operator that solves Poisson's equation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the tensor divergence
            operator. If method='auto', a suitable method is chosen
            automatically.

    Returns:
        A function that can be applied to an array of values
    """
    matrix, vector = _get_laplace_matrix(bcs)
    return make_general_poisson_solver(matrix, vector, method)


__all__ = [
    "make_laplace",
    "make_gradient",
    "make_divergence",
    "make_vector_gradient",
    "make_vector_laplace",
    "make_tensor_divergence",
    "make_poisson_solver",
]



"""
This file is used to configure the test environment when running py.test

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import matplotlib.pyplot as plt
import numpy as np
import pytest


@pytest.fixture(scope="function", autouse=True)
def suppressing():
    """ helper function adjusting message reporting for all tests """
    # raise all underflow errors
    np.seterr(all="raise", under="ignore")

    # run the actual test
    yield

    # clean up open matplotlib figures after the test
    plt.close("all")


def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)



"""
Grids define the domains on which PDEs will be solved. In particular, symmetries,
periodicities, and the discretizations are defined by the underlying grid.

We only consider regular, orthogonal grids, which are constructed from orthogonal
coordinate systems with equidistant discretizations along each axis. The dimension of
the space that the grid describes is given by the attribute :attr:`dim`. Points given in
these coordinates can be mapped to coordinates in Cartesian space using the methods
:meth:`point_to_cartesian` and its inverse. Moreover, points can be mapped to cell
indices using the methods :meth:`point_to_cell`.

.. autosummary::
   :nosignatures:

   ~cartesian.UnitGrid
   ~cartesian.CartesianGrid
   ~spherical.PolarGrid
   ~spherical.SphericalGrid
   ~cylindrical.CylindricalGrid


Inheritance structure of the classes:

.. inheritance-diagram:: pde.grids.base pde.grids.cartesian pde.grids.spherical
        pde.grids.cylindrical
   :parts: 1

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

# import all operator modules to register the operators
from . import operators
from .base import DomainError, PeriodicityError
from .cartesian import CartesianGrid, UnitGrid
from .cylindrical import CylindricalGrid
from .spherical import PolarGrid, SphericalGrid

del operators  # remove the name from the namespace
